# -*- coding: utf-8 -*-
"""ElfPatcher management and tools"""

import re
import os
import sys
import time

from udocker import is_genstr
from udocker.msg import Msg
from udocker.config import Config
from udocker.utils.uprocess import Uprocess
from udocker.utils.fileutil import FileUtil
from udocker.helper.hostinfo import HostInfo


class ElfPatcher(object):
    """Patch container executables"""

    BIN = 1
    LIB = 2
    LOADER = 4
    ABORT_ON_ERROR = 8
    ONE_SUCCESS = 16
    ONE_OUTPUT = 32

    def __init__(self, localrepo, container_id):
        self.localrepo = localrepo
        self._container_dir = \
            os.path.realpath(self.localrepo.cd_container(container_id))
        if not self._container_dir:
            raise ValueError("invalid container id")
        self._container_root = self._container_dir + "/ROOT"
        self._container_ld_so_path = self._container_dir + "/ld.so.path"
        self._container_ld_so_orig = self._container_dir + "/ld.so.orig"
        self._container_ld_libdirs = self._container_dir + "/ld.lib.dirs"
        self._container_patch_time = self._container_dir + "/patch.time"
        self._container_patch_path = self._container_dir + "/patch.path"
        self._shlib = re.compile(r"^lib\S+\.so(\.\d+)*$")
        self._uid = HostInfo.uid

    def select_patchelf(self):
        """Set patchelf executable"""
        arch = HostInfo().arch()
        image_list = []
        if arch == "amd64":
            image_list = ["patchelf-x86_64", "patchelf"]
        elif arch == "i386":
            image_list = ["patchelf-x86", "patchelf"]
        elif arch == "arm64":
            image_list = ["patchelf-arm64", "patchelf"]
        elif arch == "arm":
            image_list = ["patchelf-arm", "patchelf"]

        f_util = FileUtil(self.localrepo.bindir)
        patchelf_exec = f_util.find_file_in_dir(image_list)
        if not os.path.exists(patchelf_exec):
            Msg().err("Error: patchelf executable not found")
            sys.exit(1)

        return patchelf_exec

    def _replace(self, cmd, path):
        """Replace #f in cmd[] by path"""
        cmd_out = []
        for arg in cmd:
            if "#f" in arg:
                arg = arg.replace("#f", path)
            cmd_out.append(arg)
        return cmd_out

    def _walk_fs(self, cmd, root_path, action=BIN):
        """Execute a shell command over each executable file in a given
        dir_path, action can be ABORT_ON_ERROR, return upon first success
        ONE_SUCCESS, or return upon the first non empty output. #f is the
        placeholder for the filename.
        """
        status = ""
        for dir_path, dummy, files in os.walk(root_path):
            for f_name in files:
                try:
                    f_path = dir_path + '/' + f_name
                    if os.path.islink(f_path):
                        continue

                    if os.stat(f_path).st_uid != self._uid:
                        if action & self.ABORT_ON_ERROR:
                            return ""

                        continue

                    if ((action & self.BIN and os.access(f_path, os.X_OK)) or
                            (action & self.LIB and self._shlib.match(f_name))):
                        out = Uprocess().get_output(self._replace(cmd, f_path))
                        if out:
                            status = out

                    if action & self.ABORT_ON_ERROR and status is None:
                        return ""

                    if action & self.ONE_SUCCESS and status is not None:
                        return status

                    if action & self.ONE_OUTPUT and status:
                        return status

                except OSError:
                    pass

        return status

    def guess_elf_loader(self):
        """Search for executables and try to read the ld.so pathname"""
        patchelf_exec = self.select_patchelf()
        cmd = [patchelf_exec, "-q", "--print-interpreter", "#f"]
        for d_name in ("/bin", "/usr/bin", "/lib64"):
            elf_loader = self._walk_fs(cmd, self._container_root + d_name,
                                       self.ONE_OUTPUT | self.BIN)
            if elf_loader and ".so" in elf_loader:
                return elf_loader
        return ""

    def get_original_loader(self):
        """Get the pathname of the original ld.so"""
        if os.path.exists(self._container_ld_so_path):
            return FileUtil(self._container_ld_so_path).getdata('r').strip()
        elf_loader = self.guess_elf_loader()
        if elf_loader:
            FileUtil(self._container_ld_so_path).putdata(elf_loader, 'w')
        return elf_loader

    def get_container_loader(self):
        """Get an absolute pathname to the container ld.so"""
        elf_loader = self.get_original_loader()
        if not elf_loader:
            return ""
        elf_loader = self._container_root + "/" + elf_loader
        return elf_loader if os.path.exists(elf_loader) else ""

    def get_patch_last_path(self):
        """get last host pathname to the patched container"""
        last_path = FileUtil(self._container_patch_path).getdata('r')
        if last_path and is_genstr(last_path):
            return last_path.strip()
        return ""

    def check_container_path(self):
        """verify if path to container is ok"""
        last_path = self.get_patch_last_path()
        if last_path and last_path != self._container_dir:
            return False
        return True

    def get_patch_last_time(self):
        """get time in seconds of last full patch of container"""
        last_time = FileUtil(self._container_patch_time).getdata('r').strip()
        try:
            return str(int(last_time))
        except ValueError:
            return "0"

    def patch_binaries(self):
        """Set all executables and libs to the ld.so absolute pathname"""
        if not self.check_container_path():
            self.restore_binaries()
        last_time = '0'
        patchelf_exec = self.select_patchelf()
        elf_loader = self.get_container_loader()
        cmd = [patchelf_exec, "--set-root-prefix", self._container_root, "#f"]
        self._walk_fs(cmd, self._container_root, self.BIN | self.LIB)
        newly_set = self.guess_elf_loader()
        if newly_set == elf_loader:
            try:
                last_time = str(int(time.time()))
            except ValueError:
                pass
            futil_time = FileUtil(self._container_patch_time)
            futil_path = FileUtil(self._container_patch_path)
            return (futil_time.putdata(last_time, 'w') and
                    futil_path.putdata(self._container_dir, 'w'))
        return False

    def restore_binaries(self):
        """Restore all executables and libs to the original ld.so pathname"""
        patchelf_exec = self.select_patchelf()
        elf_loader = self.get_original_loader()
        last_path = self.get_patch_last_path()
        if last_path:
            cmd = [patchelf_exec, "--restore-root-prefix",
                   last_path + "/ROOT", "#f"]
        else:
            cmd = [patchelf_exec, "--restore-root-prefix",
                   self._container_root, "#f"]
        self._walk_fs(cmd, self._container_root, self.BIN | self.LIB)
        newly_set = self.guess_elf_loader()
        if newly_set == elf_loader:
            FileUtil(self._container_patch_path).remove()
            FileUtil(self._container_patch_time).remove()
        return newly_set == elf_loader

    def patch_ld(self, output_elf=None):
        """Patch ld.so"""
        elf_loader = self.get_container_loader()
        if FileUtil(self._container_ld_so_orig).size() == -1:
            status = FileUtil(elf_loader).copyto(self._container_ld_so_orig)
            if not status:
                return False

        ld_data = FileUtil(self._container_ld_so_orig).getdata('rb')
        if not ld_data:
            ld_data = FileUtil(elf_loader).getdata('rb')
            if not ld_data:
                return False

        nul_etc = "\x00/\x00\x00\x00\x00\x00\x00\x00\x00\x00".encode()
        nul_lib = "\x00/\x00\x00\x00".encode()
        nul_usr = "\x00/\x00\x00\x00".encode()
        etc = "\x00/etc/ld.so".encode()
        lib = "\x00/lib".encode()
        usr = "\x00/usr".encode()
        ld_data = ld_data.replace(etc, nul_etc).\
            replace(lib, nul_lib).replace(usr, nul_usr)
        ld_library_path_orig = "\x00LD_LIBRARY_PATH\x00".encode()
        ld_library_path_new = "\x00LD_LIBRARY_REAL\x00".encode()
        ld_data = ld_data.replace(ld_library_path_orig, ld_library_path_new)
        if output_elf is None:
            return bool(FileUtil(elf_loader).putdata(ld_data, 'wb'))

        return bool(FileUtil(output_elf).putdata(ld_data, 'wb'))

    def restore_ld(self):
        """Restore ld.so"""
        elf_loader = self.get_container_loader()
        futil_ldso = FileUtil(self._container_ld_so_orig)
        if futil_ldso.size() <= 0:
            Msg().err("Error: original loader not found or empty")
            return False

        if not futil_ldso.copyto(elf_loader):
            Msg().err("Error: in loader copy or file locked by other process")
            return False

        return True

    def _get_ld_config(self):
        """Get get directories from container ld.so.cache"""
        cmd = ["ldconfig", "-p", "-C",
               f"{self._container_root}/{Config.conf['ld_so_cache']}", ]
        ld_dict = {}
        ld_data = Uprocess().get_output(cmd)
        if not ld_data:
            return []
        for line in ld_data.split('\n'):
            match = re.search("([^ ]+) => ([^ ]+)", line)
            if match:
                ld_dict[self._container_root + \
                        os.path.dirname(match.group(2))] = True
        return list(ld_dict.keys())

    # pylint: disable=too-many-nested-blocks
    def _find_ld_libdirs(self, root_path=None):
        """search for library directories in container"""
        if root_path is None:
            root_path = self._container_root
        ld_list = []
        for dir_path, dummy, files in os.walk(root_path):
            for f_name in files:
                try:
                    f_path = dir_path + '/' + f_name
                    if not os.access(f_path, os.R_OK):
                        continue
                    if os.path.isfile(f_path):
                        if self._shlib.match(f_name):
                            if dir_path not in ld_list:
                                ld_list.append(dir_path)
                except OSError:
                    continue
        return ld_list

    def get_ld_libdirs(self, force=False):
        """Get ld library paths"""
        if force or not os.path.exists(self._container_ld_libdirs):
            ld_list = self._find_ld_libdirs()
            ld_str = ':'.join(ld_list)
            FileUtil(self._container_ld_libdirs).putdata(ld_str, 'w')
            return ld_list
        ld_str = FileUtil(self._container_ld_libdirs).getdata('r')
        return ld_str.split(':')

    def get_ld_library_path(self):
        """Get ld library paths"""
        ld_list = self._get_ld_config()
        ld_list.extend(self.get_ld_libdirs())
        for ld_dir in Config.conf['lib_dirs_list_essential']:
            ld_dir = self._container_root + '/' + ld_dir
            if ld_dir not in ld_list:
                ld_list.insert(0, ld_dir)
        ld_list.extend(Config.conf['lib_dirs_list_append'])
        return ':'.join(ld_list)



# -*- coding: utf-8 -*-
"""Manage setting nvidia"""

import os
import stat
import glob
import re
import shutil

from udocker.msg import Msg
from udocker.config import Config
from udocker.utils.fileutil import FileUtil
from udocker.utils.uprocess import Uprocess


class NvidiaMode(object):
    """nvidia-docker like functionality for udocker.
    Make nvidia host libraries available within udocker, this is achieved
    by copy them into the container so that the execution modes that need
    to change libraries may work properly.
    """

    def __init__(self, localrepo, container_id):
        self.localrepo = localrepo               # LocalRepository object
        self.container_id = container_id         # Container id
        self.container_dir = self.localrepo.cd_container(container_id)
        self.container_root = self.container_dir + "/ROOT"
        self._container_nvidia_set = self.container_dir + "/nvidia"
        self._nvidia_main_libs = ("libnvidia-cfg.so", "libcuda.so", )

    def _files_exist(self, cont_dst_dir, files_list):
        """Verify if files already exists"""
        for fname in files_list:
            dstname = self.container_root + '/' + cont_dst_dir + '/' + fname
            if os.path.exists(dstname):
                raise OSError("file already exists", dstname)

    def _copy_files(self, host_src_dir, cont_dst_dir, files_list, force=False):
        """copy or link file to destination creating directories as needed"""
        Msg().out("Source (host) dir ", host_src_dir, l=Msg.DBG)
        Msg().out("Destination (container) dir ", cont_dst_dir, l=Msg.DBG)
        for fname in files_list:
            srcname = host_src_dir + '/' + fname
            dstname = self.container_root + '/' + cont_dst_dir + '/' + fname
            if os.path.isfile(dstname) or os.path.islink(dstname):
                if force:
                    try:
                        os.remove(dstname)
                    except OSError:
                        Msg().err("Error: deleting nvidia file", dstname)
                else:
                    Msg().out("Info: nvidia file already exists", dstname,
                              ", use --force to overwrite", l=Msg.INF)
                    return False

            srcdir = os.path.dirname(srcname)
            dstdir = os.path.dirname(dstname)
            if not os.path.isdir(dstdir):
                try:
                    os.makedirs(dstdir)
                    os.chmod(dstdir, stat.S_IMODE(os.stat(srcdir).st_mode) |
                             stat.S_IRWXU)
                except OSError:
                    Msg().err("Error: creating nvidia dir", dstdir)

            if os.path.islink(srcname):
                linkto = os.readlink(srcname)
                os.symlink(linkto, dstname)
                Msg().out("Info: is link", srcname, "to", dstname, l=Msg.DBG)
            elif os.path.isfile(srcname):
                shutil.copy2(srcname, dstname)
                Msg().out("Info: is file", srcname, "to", dstname, l=Msg.DBG)
                try:
                    mask = stat.S_IMODE(os.stat(srcname).st_mode) | \
                                        stat.S_IWUSR | stat.S_IRUSR
                    if os.access(srcname, os.X_OK):
                        mask = mask | stat.S_IXUSR
                    os.chmod(dstname, mask)
                except (IOError, OSError) as error:
                    Msg().err("Error: change mask of nvidia file", error)
            else:
                Msg().err("Warn: nvidia file in config not found", srcname)

        Msg().out("Info: nvidia copied", srcname, "to", dstname, l=Msg.DBG)
        return True

    def _get_nvidia_libs(self, host_dir):
        """Expand the library files to include the versions"""
        lib_list = []
        for lib in Config.conf['nvi_lib_list']:
            for expanded_libs in glob.glob(host_dir + '/' + lib + '*'):
                lib_list.append(expanded_libs.replace(host_dir, ''))
        Msg().out("Info: List nvidia libs", lib_list, l=Msg.DBG)
        return lib_list

    def _find_host_dir_ldconfig(self, arch="x86-64"):
        """Find host nvidia libraries via ldconfig"""
        dir_list = set()
        ld_data = Uprocess().get_output(["ldconfig", "-p"])
        if ld_data:
            regexp = "[ |\t]%s[^ ]* .*%s.*=> (/.*)"
            for line in ld_data.split('\n'):
                for lib in self._nvidia_main_libs:
                    match = re.search(regexp % (lib, arch), line)
                    if match:
                        dir_list.add(os.path.realpath(
                            os.path.dirname(match.group(1))) + '/')
        Msg().out("Info: List nvidia libs via ldconfig", dir_list, l=Msg.DBG)
        return dir_list

    def _find_host_dir_ldpath(self, library_path):
        """Find host nvidia libraries via path"""
        dir_list = set()
        if library_path:
            for libdir in library_path.split(':'):
                for lib in self._nvidia_main_libs:
                    if glob.glob(libdir + f"/{lib}*"):
                        dir_list.add(os.path.realpath(libdir) + '/')
        Msg().out("Info: List nvidia libs via path", dir_list, l=Msg.DBG)
        return dir_list

    def _find_host_dir(self):
        """Find the location of the host nvidia libraries"""
        dir_list = set()
        library_path = ':'.join(Config.conf['lib_dirs_list_x86_64'])
        dir_list.update(self._find_host_dir_ldpath(library_path))
        dir_list.update(self._find_host_dir_ldconfig())
        library_path = os.getenv("LD_LIBRARY_PATH", "")
        dir_list.update(self._find_host_dir_ldpath(library_path))
        Msg().out("Info: Host location nvidia", dir_list, l=Msg.DBG)
        return dir_list

    def _find_cont_dir(self):
        """Find the location of the host target directory for libraries"""
        for dst_dir in ("/usr/lib/x86_64-linux-gnu", "/usr/lib64"):
            if os.path.isdir(self.container_root + '/' + dst_dir):
                Msg().out("Info: Cont. location nvidia", dst_dir, l=Msg.DBG)
                return dst_dir
        return ""

    def _installation_exists(self, nvi_host_dir_list, nvi_cont_dir):
        """Container has files from previous nvidia installation"""
        try:
            for nvi_host_dir in nvi_host_dir_list:
                lib_list = self._get_nvidia_libs(nvi_host_dir)
                self._files_exist(nvi_cont_dir, lib_list)
            self._files_exist("/etc", Config.conf['nvi_etc_list'])
            self._files_exist("/usr/bin", Config.conf['nvi_bin_list'])
            Msg().out("Info: Cont, has files from previous nvidia install")
        except OSError:
            return True
        return False

    def set_mode(self, force=False):
        """Set nvidia mode"""
        if not self.container_dir:
            Msg().err("Error: nvidia set mode container dir not found")
            return False
        nvi_host_dir_list = self._find_host_dir()
        nvi_cont_dir = self._find_cont_dir()
        if not nvi_host_dir_list:
            Msg().err("Error: host nvidia libraries not found")
            return False
        if not nvi_cont_dir:
            Msg().err("Error: destination dir for nvidia libs not found")
            return False
        if (not force) and self._installation_exists(nvi_host_dir_list,
                                                     nvi_cont_dir):
            Msg().err("Error: nvidia installation already exists"
                      ", use --force to overwrite")
            return False
        for nvi_host_dir in nvi_host_dir_list:
            lib_list = self._get_nvidia_libs(nvi_host_dir)
            self._copy_files(nvi_host_dir, nvi_cont_dir, lib_list, force)
        self._copy_files('/etc', '/etc', Config.conf['nvi_etc_list'], force)
        self._copy_files('/usr/bin', '/usr/bin', Config.conf['nvi_bin_list'],
                         force)
        FileUtil(self._container_nvidia_set).putdata("", 'w')
        Msg().out("Info: nvidia mode set")
        return True

    def get_mode(self):
        """Get nvidia mode"""
        return os.path.exists(self._container_nvidia_set)

    def get_devices(self):
        """Get list of nvidia devices related to cuda"""
        dev_list = []
        for dev in Config.conf['nvi_dev_list']:
            for expanded_devs in glob.glob(dev + '*'):
                dev_list.append(expanded_devs)
        Msg().out("Info: nvidia device list", dev_list, l=Msg.DBG)
        return dev_list



#!/usr/bin/env python
"""
udocker unit tests: PRootEngine
"""

from unittest import TestCase, main
from unittest.mock import Mock, patch
from udocker.config import Config
from udocker.engine.proot import PRootEngine


class PRootEngineTestCase(TestCase):
    """Test PRootEngine() class for containers execution."""

    def setUp(self):
        Config().getconf()
        Config().conf['hostauth_list'] = ("/etc/passwd", "/etc/group")
        Config().conf['cmd'] = "/bin/bash"
        Config().conf['cpu_affinity_exec_tools'] = \
            (["numactl", "-C", "%s", "--", ], ["taskset", "-c", "%s", ])
        Config().conf['valid_host_env'] = "HOME"
        Config().conf['username'] = "user"
        Config().conf['userhome'] = "/"
        Config().conf['oskernel'] = "4.8.13"
        Config().conf['location'] = ""

        str_local = 'udocker.container.localrepo.LocalRepository'
        self.lrepo = patch(str_local)
        self.local = self.lrepo.start()
        self.mock_lrepo = Mock()
        self.local.return_value = self.mock_lrepo

        str_exmode = 'udocker.engine.execmode.ExecutionMode'
        self.execmode = patch(str_exmode)
        self.xmode = self.execmode.start()
        self.mock_execmode = Mock()
        self.xmode.return_value = self.mock_execmode

    def tearDown(self):
        self.lrepo.stop()
        self.execmode.stop()

    @patch('udocker.engine.proot.HostInfo.oskernel')
    def test_01_init(self, mock_kernel):
        """Test01 PRootEngine() constructor."""
        mock_kernel.return_value = "4.8.13"
        prex = PRootEngine(self.local, self.xmode)
        self.assertFalse(prex.proot_noseccomp)
        self.assertEqual(prex._kernel, "4.8.13")

    @patch.object(PRootEngine, '_is_seccomp_patched')
    @patch('udocker.engine.proot.HostInfo.oskernel_isgreater')
    @patch('udocker.engine.proot.FileUtil.find_file_in_dir')
    @patch('udocker.helper.elfpatcher.os.path.exists')
    def test_02_select_proot(self, mock_exists, mock_fimage,
                             mock_isgreater, mock_issecomp):
        """Test02 PRootEngine().select_proot()."""
        Config().conf['arch'] = "amd64"
        Config().conf['proot_noseccomp'] = None
        mock_isgreater.return_value = False
        mock_fimage.return_value = "proot-4_8_0"
        mock_issecomp.return_value = False
        mock_exists.return_value = True
        prex = PRootEngine(self.local, self.xmode)
        prex.select_proot()
        self.assertFalse(prex.proot_noseccomp)

        Config().conf['proot_noseccomp'] = True
        mock_isgreater.return_value = True
        mock_fimage.return_value = "proot"
        mock_issecomp.return_value = False
        mock_exists.return_value = True
        prex = PRootEngine(self.local, self.xmode)
        prex.select_proot()
        self.assertTrue(prex.proot_noseccomp)

        Config().conf['proot_noseccomp'] = False
        mock_isgreater.return_value = True
        mock_fimage.return_value = "proot"
        mock_issecomp.return_value = False
        mock_exists.return_value = True
        prex = PRootEngine(self.local, self.xmode)
        prex.select_proot()
        self.assertFalse(prex.proot_noseccomp)

        Config().conf['proot_noseccomp'] = None
        mock_isgreater.return_value = True
        mock_fimage.return_value = "proot-x86_64-4_8_0"
        mock_issecomp.return_value = False
        mock_exists.return_value = True
        prex = PRootEngine(self.local, self.xmode)
        prex.select_proot()
        self.assertFalse(prex.proot_noseccomp)

        Config().conf['proot_noseccomp'] = False
        mock_isgreater.return_value = True
        mock_fimage.return_value = "proot-x86_64-4_8_0"
        mock_issecomp.return_value = False
        mock_exists.return_value = True
        prex = PRootEngine(self.local, self.xmode)
        prex.select_proot()
        self.assertFalse(prex.proot_noseccomp)

        Config().conf['proot_noseccomp'] = True
        mock_isgreater.return_value = True
        mock_fimage.return_value = "proot-x86_64-4_8_0"
        mock_issecomp.return_value = False
        mock_exists.return_value = True
        prex = PRootEngine(self.local, self.xmode)
        prex.select_proot()
        self.assertTrue(prex.proot_noseccomp)

    def test_03__set_uid_map(self):
        """Test03 PRootEngine()._set_uid_map()."""
        prex = PRootEngine(self.local, self.xmode)
        prex.opt["uid"] = "0"
        status = prex._set_uid_map()
        self.assertEqual(status, ['-0'])

        prex = PRootEngine(self.local, self.xmode)
        prex.opt["uid"] = "1000"
        prex.opt["gid"] = "1001"
        status = prex._set_uid_map()
        self.assertEqual(status, ['-i', '1000:1001'])

    def test_04__create_mountpoint(self):
        """Test04 PRootEngine()._create_mountpoint()."""
        prex = PRootEngine(self.local, self.xmode)
        status = prex._create_mountpoint("", "")
        self.assertTrue(status)

    def test_05__get_volume_bindings(self):
        """Test05 PRootEngine()._get_volume_bindings()."""
        prex = PRootEngine(self.local, self.xmode)
        prex.opt["vol"] = ()
        status = prex._get_volume_bindings()
        self.assertEqual(status, [])

        prex = PRootEngine(self.local, self.xmode)
        prex.opt["vol"] = ("/tmp", "/bbb",)
        status = prex._get_volume_bindings()
        self.assertEqual(status, ['-b', '/tmp:/tmp', '-b', '/bbb:/bbb'])

    @patch('udocker.engine.proot.ExecutionEngineCommon._get_portsmap')
    def test_06__get_network_map(self, mock_pmap):
        """Test06 PRootEngine()._get_network_map()."""
        mock_pmap.return_value = {}
        prex = PRootEngine(self.local, self.xmode)
        prex.opt['netcoop'] = False
        status = prex._get_network_map()
        self.assertEqual(status, [])

        mock_pmap.return_value = {}
        prex = PRootEngine(self.local, self.xmode)
        prex.opt['netcoop'] = True
        status = prex._get_network_map()
        self.assertEqual(status, ['-n'])

        mock_pmap.return_value = {80: 8080, 443: 8443}
        prex = PRootEngine(self.local, self.xmode)
        prex.opt['netcoop'] = True
        status = prex._get_network_map()
        self.assertEqual(status, ['-p', '80:8080', '-p', '443:8443', '-n'])

    @patch('udocker.engine.proot.os.environ.update')
    @patch.object(PRootEngine, '_get_network_map')
    @patch('udocker.engine.nvidia.Msg')
    @patch('udocker.engine.proot.HostInfo.oskernel_isgreater')
    @patch.object(PRootEngine, '_run_banner')
    @patch.object(PRootEngine, '_run_env_cleanup_dict')
    @patch.object(PRootEngine, '_set_uid_map')
    @patch.object(PRootEngine, '_get_volume_bindings')
    @patch.object(PRootEngine, '_set_cpu_affinity')
    @patch.object(PRootEngine, '_run_env_set')
    @patch.object(PRootEngine, 'select_proot')
    @patch('udocker.engine.proot.os.getenv')
    @patch('udocker.engine.proot.subprocess.call')
    @patch('udocker.engine.proot.ExecutionEngineCommon._run_init')
    def test_07_run(self, mock_run_init, mock_call, mock_getenv, mock_sel_proot,
                    mock_run_env_set,
                    mock_set_cpu_aff, mock_get_vol_bind, mock_set_uid_map,
                    mock_env_cleanup_dict, mock_run_banner, mock_isgreater,
                    mock_msg, mock_netmap, mock_envupd):
        """Test07 PRootEngine().run()."""
        mock_run_init.return_value = False
        prex = PRootEngine(self.local, self.xmode)
        status = prex.run("CONTAINERID")
        self.assertEqual(status, 2)

        Config().conf['proot_killonexit'] = False
        mock_run_init.return_value = True
        mock_sel_proot.return_value = None
        mock_isgreater.return_value = "2.9.0"
        mock_run_env_set.return_value = None
        mock_msg.level = 5
        mock_set_cpu_aff.return_value = []
        mock_get_vol_bind.return_value = ['-b', '/tmp:/tmp']
        mock_set_uid_map.return_value = ['-i', '1000:1001']
        mock_netmap.return_value = ['-p', '80:8080', '-n']
        mock_env_cleanup_dict.return_value = None
        mock_run_banner.return_value = None
        mock_call.return_value = 5
        mock_getenv.return_value = False
        mock_envupd.return_value = None
        prex = PRootEngine(self.local, self.xmode)
        prex.opt["kernel"] = "5.4.0"
        status = prex.run("CONTAINERID")
        self.assertEqual(status, 5)
        self.assertTrue(mock_run_init.called)
        self.assertTrue(mock_call.called)
        self.assertTrue(mock_envupd.called)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
udocker unit tests: SingularityEngine
"""

from unittest import TestCase, main
from unittest.mock import Mock, patch
from udocker.config import Config
from udocker.engine.singularity import SingularityEngine


class SingularityEngineTestCase(TestCase):
    """Test SingularityEngine() class for containers execution."""

    def setUp(self):
        Config().getconf()

        str_local = 'udocker.container.localrepo.LocalRepository'
        self.lrepo = patch(str_local)
        self.local = self.lrepo.start()
        self.mock_lrepo = Mock()
        self.local.return_value = self.mock_lrepo

        str_exmode = 'udocker.engine.execmode.ExecutionMode'
        self.execmode = patch(str_exmode)
        self.xmode = self.execmode.start()
        self.mock_execmode = Mock()
        self.xmode.return_value = self.mock_execmode

    def tearDown(self):
        self.lrepo.stop()
        self.execmode.stop()

    def test_01_init(self):
        """Test01 SingularityEngine() constructor."""
        sing = SingularityEngine(self.local, self.xmode)
        self.assertEqual(sing.executable, None)
        self.assertEqual(sing.execution_id, None)

    @patch('udocker.helper.elfpatcher.os.path.exists')
    @patch('udocker.engine.singularity.HostInfo.arch')
    @patch('udocker.engine.singularity.FileUtil.find_file_in_dir')
    @patch('udocker.engine.singularity.FileUtil.find_exec')
    def test_02_select_singularity(self, mock_findexe, mock_find,
                                   mock_arch, mock_exists):
        """Test02 SingularityEngine().select_singularity()."""
        Config.conf['use_singularity_executable'] = ""
        mock_findexe.return_value = "/bin/runc-arm"
        mock_exists.return_value = True
        sing = SingularityEngine(self.local, self.xmode)
        sing.select_singularity()
        self.assertTrue(mock_findexe.called)

        Config.conf['use_singularity_executable'] = "UDOCKER"
        mock_arch.return_value = "amd64"
        mock_find.return_value = "singularity-x86_64"
        mock_exists.return_value = True
        sing = SingularityEngine(self.local, self.xmode)
        sing.select_singularity()
        self.assertTrue(mock_arch.called)
        self.assertTrue(mock_find.called)

        Config.conf['use_singularity_executable'] = "UDOCKER"
        mock_arch.return_value = "i386"
        mock_find.return_value = "singularity-x86"
        mock_exists.return_value = True
        sing = SingularityEngine(self.local, self.xmode)
        sing.select_singularity()
        self.assertTrue(mock_arch.called)
        self.assertTrue(mock_find.called)

    @patch('udocker.engine.singularity.NixAuthentication.get_home')
    @patch('udocker.engine.singularity.os.path.isdir')
    def test_03__get_volume_bindings(self, mock_isdir, mock_gethome):
        """Test03 SingularityEngine()._get_volume_bindings()."""
        res_exp = ['-B', '/dir1:/dir1',
                   '-B', '/home/user:/home/user',
                   '-B', '/tmp:/tmp',
                   '-B', '/var/tmp:/var/tmp']
        mock_gethome.return_value = "/home/user"
        mock_isdir.side_effect = [True, True]
        sing = SingularityEngine(self.local, self.xmode)
        sing.opt = dict()
        sing.opt["vol"] = ["/dir1", "/home/user"]
        status = sing._get_volume_bindings()
        self.assertEqual(status, res_exp)
        self.assertTrue(mock_gethome.called)
        self.assertTrue(mock_isdir.call_count, 2)

        res_exp = ['-B', '/home/user:/home/user',
                   '-B', '/tmp:/tmp',
                   '-B', '/var/tmp:/var/tmp']
        mock_gethome.return_value = "/home/user"
        mock_isdir.side_effect = [True, True, True]
        sing = SingularityEngine(self.local, self.xmode)
        sing.opt = dict()
        sing.opt["vol"] = ["/home/user", "/tmp", "/var/tmp"]
        status = sing._get_volume_bindings()
        self.assertEqual(status, res_exp)
        self.assertTrue(mock_gethome.called)
        self.assertTrue(mock_isdir.call_count, 3)

    def test_04__singularity_env_get(self):
        """Test04 SingularityEngine()._singularity_env_get()."""
        res_exp = {"SINGULARITYENV_AA": "aa", "SINGULARITYENV_BB": "bb"}
        sing = SingularityEngine(self.local, self.xmode)
        sing.opt = dict()
        sing.opt["env"] = [("AA", "aa"), ("BB", "bb")]
        status = sing._singularity_env_get()
        self.assertEqual(status, res_exp)

    @patch('udocker.engine.singularity.FileUtil.mkdir')
    def test_05__make_container_dirs(self, mock_mkdir):
        """Test05 SingularityEngine()._make_container_directories()."""
        mock_mkdir.side_effect = [None, None, None, None, None, None]
        sing = SingularityEngine(self.local, self.xmode)
        sing._make_container_directories()
        self.assertTrue(mock_mkdir.call_count, 6)

    @patch('udocker.engine.singularity.Msg')
    def test_06__run_invalid_options(self, mock_msg):
        """Test06 SingularityEngine()._run_invalid_options()."""
        mock_msg.level = 0
        sing = SingularityEngine(self.local, self.xmode)
        sing.opt['netcoop'] = False
        sing.opt['portsmap'] = True
        sing._run_invalid_options()
        self.assertTrue(mock_msg.called)

    @patch.object(SingularityEngine, '_has_option')
    @patch('udocker.engine.singularity.NixAuthentication')
    @patch('udocker.engine.singularity.HostInfo.username')
    @patch('udocker.engine.singularity.Msg')
    def test_07__run_as_root(self, mock_msg, mock_uname,
                             mock_nixauth, mock_hasopt):
        """Test07 SingularityEngine()._run_as_root()."""
        mock_msg.level = 0
        mock_uname.return_value = "u1"
        sing = SingularityEngine(self.local, self.xmode)
        sing.opt = dict()
        status = sing._run_as_root()
        self.assertFalse(status)
        self.assertTrue(mock_uname.called)

        mock_msg.level = 0
        mock_uname.return_value = "u1"
        sing = SingularityEngine(self.local, self.xmode)
        sing.opt = dict()
        sing.opt["user"] = "u1"
        status = sing._run_as_root()
        self.assertFalse(status)
        self.assertTrue(mock_uname.called)

        mock_msg.level = 0
        mock_uname.return_value = "u1"
        sing = SingularityEngine(self.local, self.xmode)
        sing.opt = dict()
        sing.opt["user"] = "u2"
        sing.opt["uid"] = "1000"
        status = sing._run_as_root()
        self.assertFalse(status)
        self.assertTrue(mock_uname.called)

        mock_msg.level = 0
        mock_uname.return_value = "u1"
        mock_hasopt.return_value = True
        mock_nixauth.return_value.user_in_subuid.return_value = True
        mock_nixauth.return_value.user_in_subgid.return_value = True
        sing = SingularityEngine(self.local, self.xmode)
        sing.opt = dict()
        sing.opt["user"] = "root"
        sing.opt["uid"] = "0"
        status = sing._run_as_root()
        self.assertTrue(status)
        self.assertTrue(mock_hasopt.called)
        self.assertTrue(mock_nixauth.return_value.user_in_subuid.called)
        self.assertTrue(mock_nixauth.return_value.user_in_subgid.called)

    @patch.object(SingularityEngine, '_run_banner')
    @patch.object(SingularityEngine, '_run_env_cleanup_dict')
    @patch.object(SingularityEngine, '_get_volume_bindings')
    @patch.object(SingularityEngine, '_run_env_set')
    @patch.object(SingularityEngine, '_run_as_root')
    @patch.object(SingularityEngine, 'select_singularity')
    @patch.object(SingularityEngine, '_make_container_directories')
    @patch.object(SingularityEngine, '_run_invalid_options')
    @patch.object(SingularityEngine, '_run_init')
    @patch('udocker.engine.singularity.subprocess.call')
    @patch('udocker.engine.singularity.Unique.uuid')
    @patch('udocker.engine.singularity.FileBind')
    @patch('udocker.engine.singularity.os.path.isdir')
    def test_08_run(self, mock_isdir, mock_fbind, mock_uuid, mock_call,
                    mock_rinit, mock_rinvopt, mock_mkcontdir,
                    mock_selsing, mock_runroot, mock_renvset, mock_getvolbind,
                    mock_renvclean, mock_rbann):
        """Test08 SingularityEngine().run()."""
        mock_isdir.return_value = True
        mock_fbind.return_value.container_orig_dir.return_value = "/d1"
        mock_fbind.return_value.restore.return_value = None
        mock_rinit.return_value = False
        sing = SingularityEngine(self.local, self.xmode)
        status = sing.run("CONTAINERID")
        self.assertEqual(status, 2)

        mock_isdir.return_value = True
        mock_fbind.return_value.container_orig_dir.return_value = "/d1"
        mock_fbind.return_value.restore.return_value = None
        mock_rinit.return_value = "/cont/"
        mock_rinvopt.return_value = None
        mock_mkcontdir.return_value = None
        mock_selsing.return_value = None
        mock_runroot.return_value = None
        mock_renvset.return_value = None
        self.local.bindir = "/cont/bin"
        mock_getvolbind.return_value = ['-B', '/home/user:/home/user',
                                        '-B', '/tmp:/tmp',
                                        '-B', '/var/tmp:/var/tmp']
        mock_uuid.return_value = "EXECUTION_ID"
        mock_renvclean.return_value = None
        mock_rbann.return_value = None
        mock_call.return_value = 0
        sing = SingularityEngine(self.local, self.xmode)
        sing.executable = "/bin/sing"
        status = sing.run("CONTAINERID")
        self.assertEqual(status, 0)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Classes for cURL management and tools"""

import os
import sys
import json

from udocker import is_genstr
from udocker.config import Config
from udocker.msg import Msg
from udocker.utils.fileutil import FileUtil
from udocker.utils.uprocess import Uprocess

try:
    import pycurl
except ImportError:
    pass

# if Python 3
# pylint: disable=import-error
if sys.version_info[0] >= 3:
    from io import BytesIO as strio
else:
    from StringIO import StringIO as strio


class CurlHeader(object):
    """An http header parser to be used with PyCurl
    Allows to retrieve the header fields and the status.
    Allows to obtain just the header by stopping the
    download (returning -1) e.g to get the content length
    this is useful if the RESTful interface or web server
    does not implement HEAD.
    """

    def __init__(self):
        self.sizeonly = False
        self.data = {}
        self.data["X-ND-HTTPSTATUS"] = ""
        self.data["X-ND-CURLSTATUS"] = ""

    def write(self, buff):
        """Write is called by Curl()"""
        if not isinstance(buff, str):
            buff = buff.decode()
        pair = buff.split(":", 1)
        if len(pair) == 2:
            key = pair[0].strip().lower()
            if key:
                self.data[key] = pair[1].strip()
        elif pair[0].startswith("HTTP/"):
            self.data["X-ND-HTTPSTATUS"] = buff.strip()
        elif (self.sizeonly and
              pair[0].strip() == "" and
              "location" not in self.data):
            return -1
        return None

    def setvalue_from_file(self, in_filename):
        """Load header content from file instead of from Curl()
        Alternative to write() to be used with the curl executable
        version.
        """
        try:
            infile = open(in_filename, 'r', encoding='utf-8')
        except (IOError, OSError):
            return False
        for line in infile:
            self.write(line)
        infile.close()
        return True

    def getvalue(self):
        """Return the curl data buffer"""
        return str(self.data)

    def __str__(self):
        """Return a string representation"""
        return str(self.data)


class GetURL(object):
    """File downloader using PyCurl or a curl cli executable"""

    def __init__(self):
        """Load configuration common to the implementations"""
        self.timeout = Config.conf['timeout']
        self.ctimeout = Config.conf['ctimeout']
        self.download_timeout = Config.conf['download_timeout']
        self.agent = Config.conf['http_agent']
        self.http_proxy = Config.conf['http_proxy']
        self.cache_support = False
        self.insecure = Config.conf['http_insecure']
        self._curl_exec = Config.conf['use_curl_executable']
        self._select_implementation()

    # pylint: disable=locally-disabled
    # pylint: disable=protected-access
    def _select_implementation(self):
        """Select which implementation to use"""
        if GetURLpyCurl().is_available() and not self._curl_exec:
            self._geturl = GetURLpyCurl()
            self.cache_support = True
            Msg().out("Info: using pycurl", l=Msg.DBG)
        elif GetURLexeCurl().is_available():
            self._geturl = GetURLexeCurl()
            Msg().out("Info: using curl executable", self._geturl._curl_exec,
                      l=Msg.DBG)
        else:
            Msg().err("Error: need curl or pycurl to perform downloads")
            raise NameError('need curl or pycurl')

    def get_content_length(self, hdr):
        """Get content length from the http header"""
        try:
            return int(hdr.data["content-length"])
        except (ValueError, TypeError):
            return -1

    def set_insecure(self, bool_value=True):
        """Use insecure downloads no SSL verification"""
        self.insecure = bool_value
        self._geturl.insecure = bool_value

    def set_proxy(self, http_proxy):
        """Specify a socks http proxy"""
        self.http_proxy = http_proxy
        self._geturl.http_proxy = http_proxy

    def get(self, *args, **kwargs):
        """Get URL using selected implementation
        Example:
            get(url, ctimeout=5, timeout=5, header=[]):
        """
        if len(args) != 1:
            raise TypeError('wrong number of arguments')
        return self._geturl.get(*args, **kwargs)

    def post(self, *args, **kwargs):
        """POST using selected implementation"""
        if len(args) != 2:
            raise TypeError('wrong number of arguments')
        kwargs["post"] = args[1]
        return self._geturl.get(args[0], **kwargs)

    def get_status_code(self, status_line):
        """
        Get http status code from http status line.
        Status-Line = HTTP-Version Status-Code Reason-Phrase CRLF
        """
        try:
            return int(status_line.split(' ')[1])
        except ValueError:
            return 400
        except IndexError:
            return 404


class GetURLpyCurl(GetURL):
    """Downloader implementation using PyCurl"""

    def __init__(self):
        GetURL.__init__(self)
        self._url = None

    def is_available(self):
        """Can we use this approach for download"""
        try:
            dummy = pycurl.Curl()
        except (NameError, AttributeError):
            return False
        return True

    def _select_implementation(self):
        """Override the parent class method"""
        return

    def _set_defaults(self, pyc, hdr):
        """Set options for pycurl"""
        if self.insecure:
            pyc.setopt(pyc.SSL_VERIFYPEER, 0)
            pyc.setopt(pyc.SSL_VERIFYHOST, 0)

        pyc.setopt(pyc.FOLLOWLOCATION, False)
        pyc.setopt(pyc.FAILONERROR, False)
        pyc.setopt(pyc.NOPROGRESS, True)
        pyc.setopt(pyc.HEADERFUNCTION, hdr.write)
        pyc.setopt(pyc.USERAGENT, self.agent)
        pyc.setopt(pyc.CONNECTTIMEOUT, self.ctimeout)
        pyc.setopt(pyc.TIMEOUT, self.timeout)
        pyc.setopt(pyc.PROXY, self.http_proxy)
        if Msg.level >= Msg.VER:
            pyc.setopt(pyc.VERBOSE, True)
        else:
            pyc.setopt(pyc.VERBOSE, False)
        self._url = ""

    def _mkpycurl(self, pyc, hdr, buf, *args, **kwargs):
        """Prepare curl command line according to invocation options"""
        self._url = str(args[0])
        pyc.setopt(pycurl.URL, self._url)
        if "follow" in kwargs:
            pyc.setopt(pyc.FOLLOWLOCATION, kwargs["follow"])
        if "post" in kwargs:
            pyc.setopt(pyc.POST, 1)
            pyc.setopt(pyc.HTTPHEADER, ['Content-Type: application/json'])
            pyc.setopt(pyc.POSTFIELDS, json.dumps(kwargs["post"]))
        if "sizeonly" in kwargs:
            hdr.sizeonly = True
        if "proxy" in kwargs and kwargs["proxy"]:
            pyc.setopt(pyc.PROXY, pyc.USERAGENT, kwargs["proxy"])
        if "ctimeout" in kwargs:
            pyc.setopt(pyc.CONNECTTIMEOUT, kwargs["ctimeout"])
        if "header" in kwargs:  # avoid known pycurl bug
            clean_header_list = []
            for header_item in kwargs["header"]:
                if str(header_item).startswith("Authorization: Bearer"):
                    if "Signature=" in self._url:
                        continue
                    if "redirect" in kwargs:
                        continue
                clean_header_list.append(str(header_item))
            pyc.setopt(pyc.HTTPHEADER, clean_header_list)
        if "v" in kwargs:
            pyc.setopt(pyc.VERBOSE, kwargs["v"])
        if "nobody" in kwargs:
            pyc.setopt(pyc.NOBODY, kwargs["nobody"])  # header only no content
        if "timeout" in kwargs:
            pyc.setopt(pyc.TIMEOUT, kwargs["timeout"])
        if "ofile" in kwargs:
            output_file = kwargs["ofile"]
            pyc.setopt(pyc.TIMEOUT, self.download_timeout)
            openflags = "wb"
            if "resume" in kwargs and kwargs["resume"]:
                pyc.setopt(pyc.RESUME_FROM, FileUtil(output_file).size())
                openflags = "ab"
            try:
                filep = open(output_file, openflags)
            except(IOError, OSError):
                Msg().err(f"Error: opening download file: {output_file}")
                raise
            pyc.setopt(pyc.WRITEDATA, filep)
        else:
            filep = None
            output_file = ""
            pyc.setopt(pyc.WRITEFUNCTION, buf.write)
        hdr.data["X-ND-CURLSTATUS"] = 0
        return(output_file, filep)

    def get(self, *args, **kwargs):
        """http get implementation using the PyCurl"""
        hdr = CurlHeader()
        buf = strio()
        pyc = pycurl.Curl()
        self._set_defaults(pyc, hdr)
        try:
            (output_file, filep) = \
                    self._mkpycurl(pyc, hdr, buf, *args, **kwargs)
            Msg().out("curl url: ", self._url, l=Msg.DBG)
            Msg().out("curl arg: ", kwargs, l=Msg.DBG)
            pyc.perform()     # call pyculr
        except(IOError, OSError):
            return (None, None)
        except pycurl.error as error:
            # pylint: disable=unbalanced-tuple-unpacking
            errno, errstr = error.args
            hdr.data["X-ND-CURLSTATUS"] = errno
            if not hdr.data["X-ND-HTTPSTATUS"]:
                hdr.data["X-ND-HTTPSTATUS"] = errstr
        status_code = self.get_status_code(hdr.data["X-ND-HTTPSTATUS"])
        if "header" in kwargs:
            hdr.data["X-ND-HEADERS"] = kwargs["header"]
        if status_code == 401: # needs authentication
            pass
        elif 300 <= status_code <= 308: # redirect
            pass
        elif "ofile" in kwargs:
            filep.close()
            if status_code == 206 and "resume" in kwargs:
                pass
            elif status_code == 416 and "resume" in kwargs:
                kwargs["resume"] = False
                (hdr, buf) = self.get(self._url, **kwargs)
            elif status_code != 200:
                Msg().err("Error: in download: " + str(
                    hdr.data["X-ND-HTTPSTATUS"]))
                FileUtil(output_file).remove()
        return (hdr, buf)


class GetURLexeCurl(GetURL):
    """Downloader implementation using curl cli executable"""

    def __init__(self):
        GetURL.__init__(self)
        self._opts = None
        self._files = None

    def is_available(self):
        """Can we use this approach for download"""
        return bool(FileUtil("curl").find_exec())

    def _select_implementation(self):
        """Override the parent class method"""
        return

    def _set_defaults(self):
        """Set defaults for curl command line options"""
        self._opts = {
            "insecure": [],
            "header": [],
            "verbose": [],
            "nobody": [],
            "proxy": [],
            "resume": [],
            "ctimeout": ["--connect-timeout", str(self.ctimeout)],
            "timeout": ["-m", str(self.timeout)],
            "other": ["-s", "-q", "-S"]
        }
        if self.insecure:
            self._opts["insecure"] = ["-k"]
        if Msg().level > Msg.DBG:
            self._opts["verbose"] = ["-v"]
        self._files = {
            "url":  "",
            "error_file": FileUtil("execurl_err").mktmp(),
            "output_file": FileUtil("execurl_out").mktmp(),
            "header_file": FileUtil("execurl_hdr").mktmp()
        }

    def _mkcurlcmd(self, *args, **kwargs):
        """Prepare curl command line according to invocation options"""
        self._files["url"] = str(args[0])
        if "follow" in kwargs and kwargs["follow"]:
            self._opts["follow"] = ["-L"]
        if "post" in kwargs:
            self._opts["post"] = ["-X", "POST", "-H",
                                  "Content-Type: application/json"]
            self._opts["post"] += ["-d", json.dumps(kwargs["post"])]
        if "ctimeout" in kwargs:
            self._opts["ctimeout"] = ["--connect-timeout",
                                      str(kwargs["ctimeout"])]
        if "timeout" in kwargs:
            self._opts["timeout"] = ["-m", str(kwargs["timeout"])]
        if "proxy" in kwargs and kwargs["proxy"]:
            self._opts["proxy"] = ["--proxy", str(kwargs["proxy"])]
        elif self.http_proxy:
            self._opts["proxy"] = ["--proxy", self.http_proxy]
        if "header" in kwargs:
            for header_item in kwargs["header"]:
                if str(header_item).startswith("Authorization: Bearer"):
                    if "Signature=" in self._files["url"]:
                        continue
                    if "redirect" in kwargs:
                        continue
                self._opts["header"] += ["-H", str(header_item)]
        if 'v' in kwargs and kwargs['v']:
            self._opts["verbose"] = ["-v"]
        if "nobody" in kwargs and kwargs["nobody"]:
            self._opts["nobody"] = ["--head"]
        if "ofile" in kwargs:
            FileUtil(self._files["output_file"]).remove()
            self._files["output_file"] = kwargs["ofile"] + ".tmp"
            self._opts["timeout"] = ["-m", str(self.download_timeout)]
            if "resume" in kwargs and kwargs["resume"]:
                self._opts["resume"] = ["-C", "-"]
        cmd = ["curl"]
        if self._curl_exec and is_genstr(self._curl_exec):
            cmd = [self._curl_exec]
        for opt in self._opts.values():
            cmd += opt
        cmd.extend(["-D", self._files["header_file"], "-o",
                    self._files["output_file"], "--stderr",
                    self._files["error_file"], self._files["url"]])
        return cmd

    def get(self, *args, **kwargs):
        """http get implementation using the curl cli executable"""
        hdr = CurlHeader()
        buf = strio()
        self._set_defaults()
        cmd = self._mkcurlcmd(*args, **kwargs)
        status = Uprocess().call(cmd, close_fds=True, stderr=Msg.chlderr,
                                 stdout=Msg.chlderr) # call curl
        hdr.setvalue_from_file(self._files["header_file"])
        hdr.data["X-ND-CURLSTATUS"] = status
        if status:
            err_down = str(FileUtil(self._files["error_file"]).getdata('r'))
            Msg().err(f"Error: in download: {err_down}")
            FileUtil(self._files["output_file"]).remove()
            return (hdr, buf)
        status_code = self.get_status_code(hdr.data["X-ND-HTTPSTATUS"])
        if "header" in kwargs:
            hdr.data["X-ND-HEADERS"] = kwargs["header"]
        if status_code == 401: # needs authentication
            pass
        elif 300 <= status_code <= 308: # redirect
            pass
        elif "ofile" in kwargs:
            if status_code == 206 and "resume" in kwargs:
                os.rename(self._files["output_file"], kwargs["ofile"])
            elif status_code == 416:
                if "resume" in kwargs:
                    kwargs["resume"] = False
                (hdr, buf) = self.get(self._files["url"], **kwargs)
            elif status_code != 200:
                Msg().err("Error: in download: ", str(
                    hdr.data["X-ND-HTTPSTATUS"]), ": ", str(status))
                FileUtil(self._files["output_file"]).remove()
            else:  # OK downloaded
                os.rename(self._files["output_file"], kwargs["ofile"])
        if "ofile" not in kwargs:
            try:
                buf = strio(open(self._files["output_file"], 'rb').read())
            except(IOError, OSError):
                Msg().err("Error: reading curl output file to buffer")
            FileUtil(self._files["output_file"]).remove()
        FileUtil(self._files["error_file"]).remove()
        FileUtil(self._files["header_file"]).remove()
        return (hdr, buf)



# -*- coding: utf-8 -*-
"""File utilities"""

import os
import sys
import stat
import re

from udocker import is_genstr
from udocker.msg import Msg
from udocker.config import Config
from udocker.helper.unique import Unique
from udocker.helper.hostinfo import HostInfo
from udocker.utils.uprocess import Uprocess
from udocker.utils.uvolume import Uvolume


class FileUtil(object):
    """Some utilities to manipulate files"""

    tmptrash = {}
    safe_prefixes = []
    orig_umask = None

    def __init__(self, filename=None):
        self._tmpdir = Config.conf['tmpdir']
        self.orig_filename = filename
        if filename == "-":
            self.filename = "-"
            self.basename = "-"
            return
        try:
            self.filename = os.path.abspath(filename)
            self.basename = os.path.basename(self.filename)
        except (AttributeError, TypeError):
            self.filename = filename
            self.basename = filename
        self._register_prefix(self._tmpdir)

    def _register_prefix(self, prefix):
        """Register directory prefixes where remove() is allowed"""
        prefix = os.path.realpath(prefix)
        if prefix not in FileUtil.safe_prefixes:
            filename = prefix
            if os.path.isdir(filename) and not filename.endswith('/'):
                FileUtil.safe_prefixes.append(filename + '/')
                FileUtil.safe_prefixes.append(os.path.realpath(filename) + '/')
            else:
                FileUtil.safe_prefixes.append(filename)
                FileUtil.safe_prefixes.append(os.path.realpath(filename))

    def register_prefix(self):
        """Register directory prefixes where remove() is allowed"""
        self._register_prefix(self.filename)

    def umask(self, new_umask=None):
        """Set umask"""
        if new_umask is not None:
            try:
                old_umask = os.umask(new_umask)
            except (TypeError, ValueError):
                return False
            if FileUtil.orig_umask is None:
                FileUtil.orig_umask = old_umask
        else:
            try:
                os.umask(FileUtil.orig_umask)
            except (TypeError, ValueError):
                return False
        return True

    def mktmp(self):
        """Generate a temporary filename"""
        while True:
            tmp_file = self._tmpdir + '/' + \
                Unique().filename(self.basename)
            if not os.path.exists(tmp_file):
                FileUtil.tmptrash[tmp_file] = True
                self.filename = tmp_file
                return tmp_file

    def mkdir(self):
        """Create directory"""
        try:
            os.makedirs(self.filename)
        except (OSError, IOError, AttributeError):
            return False
        return True

    def rmdir(self):
        """Remove an empty directory"""
        try:
            os.rmdir(self.filename)
        except (OSError, IOError, AttributeError):
            return False
        return True

    def mktmpdir(self):
        """Create temporary directory"""
        dirname = self.mktmp()
        if FileUtil(dirname).mkdir():
            return dirname
        return None

    def uid(self):
        """Get the file owner user id"""
        try:
            return os.lstat(self.filename).st_uid
        except (IOError, OSError):
            return -1

    def _is_safe_prefix(self, filename):
        """Check if file prefix falls under valid prefixes"""
        filename = os.path.realpath(filename)
        if os.path.isdir(filename):
            filename += '/'
        for safe_prefix in FileUtil.safe_prefixes:
            if filename.startswith(safe_prefix):
                return True
            if filename.startswith(os.path.realpath(safe_prefix)):
                return True
        return False

    def chown(self, uid=0, gid=0, recursive=False):
        """Change ownership of file or directory"""
        try:
            if recursive:
                for dir_path, dirs, files in os.walk(self.filename):
                    for f_name in dirs + files:
                        os.lchown(dir_path + '/' + f_name, uid, gid)
            else:
                os.lchown(self.filename, uid, gid)
        except OSError:
            return False
        return True

    def rchown(self, uid=0, gid=0):
        """Change ownership recursively recursively"""
        return self.chown(uid, gid, recursive=True)

    def _chmod(self, filename, filemode=0o600, dirmode=0o700, mask=0o755):
        """chmod file or directory"""
        try:
            filestat = os.lstat(filename).st_mode
            if stat.S_ISREG(filestat) and filemode:
                mode = (stat.S_IMODE(filestat) & mask) | filemode
                os.chmod(filename, mode)
            elif stat.S_ISDIR(filestat) and dirmode:
                mode = (stat.S_IMODE(filestat) & mask) | dirmode
                os.chmod(filename, mode)
            elif stat.S_ISLNK(filestat) and filemode:
                pass
            elif filemode:
                mode = (stat.S_IMODE(filestat) & mask) | filemode
                os.chmod(filename, mode)
        except OSError:
            Msg().err("Error: changing permissions of:", filename, l=Msg.VER)
            raise OSError

    def chmod(self, filemode=0o600, dirmode=0o700, mask=0o755, recursive=False):
        """chmod directory recursively"""
        try:
            if recursive:
                for dir_path, dirs, files in os.walk(self.filename):
                    for f_name in files:
                        self._chmod(dir_path + '/' + f_name,
                                    filemode, None, mask)
                    for f_name in dirs:
                        self._chmod(dir_path + '/' + f_name,
                                    None, dirmode, mask)
            self._chmod(self.filename, filemode, dirmode, mask)
        except OSError:
            return False
        return True

    def rchmod(self, filemode=0o600, dirmode=0o700, mask=0o755):
        """chmod directory recursively"""
        self.chmod(filemode, dirmode, mask, True)

    def _removedir(self):
        """Delete directory recursively"""
        try:
            for dir_path, dirs, files in os.walk(self.filename, topdown=False,
                                                 followlinks=False):
                for f_name in files:
                    f_path = dir_path + '/' + f_name
                    if not os.path.islink(f_path):
                        os.chmod(f_path, stat.S_IWUSR | stat.S_IRUSR)
                    os.unlink(f_path)
                for f_name in dirs:
                    f_path = dir_path + '/' + f_name
                    if os.path.islink(f_path):
                        os.unlink(f_path)
                        continue
                    os.chmod(f_path, stat.S_IWUSR | stat.S_IRUSR | stat.S_IXUSR)
                    os.rmdir(f_path)
            os.chmod(self.filename, stat.S_IWUSR | stat.S_IRUSR | stat.S_IXUSR)
            os.rmdir(self.filename)
        except OSError:
            Msg().err("Error: removing:", self.filename, l=Msg.VER)
            return False
        return True

    def remove(self, force=False, recursive=False):
        """Delete files or directories"""
        if not os.path.lexists(self.filename):
            pass
        elif self.filename.count("/") < 2:
            Msg().err("Error: delete pathname too short: ", self.filename)
            return False
        elif self.uid() != HostInfo.uid:
            Msg().err("Error: delete not owner: ", self.filename)
            return False
        elif (not force) and (not self._is_safe_prefix(self.filename)):
            Msg().err("Error: delete outside of directory tree: ",
                      self.filename)
            return False
        elif os.path.isfile(self.filename) or os.path.islink(self.filename):
            try:
                os.remove(self.filename)
            except (IOError, OSError):
                Msg().err("Error: deleting file: ", self.filename)
                return False
        elif os.path.isdir(self.filename):
            if recursive:
                status = self._removedir()
            else:
                status = self.rmdir()
            if not status:
                Msg().err("Error: deleting directory: ", self.filename)
                return False
        if self.filename in dict(FileUtil.tmptrash):
            del FileUtil.tmptrash[self.filename]
        return True

    def verify_tar(self):
        """Verify a tar file: tar tvf file.tar"""
        if not os.path.isfile(self.filename):
            return False
        verbose = ''
        if Msg.level >= Msg.VER:
            verbose = 'v'
        cmd = ["tar", "t" + verbose + "f", self.filename]
        if Uprocess().call(cmd, stderr=Msg.chlderr, stdout=Msg.chlderr,
                           close_fds=True):
            return False
        return True

    def tar(self, tarfile, sourcedir=None):
        """Create a tar file for a given sourcedir"""
        #cmd += r" --xform 's:^\./::' "
        if sourcedir is None:
            sourcedir = self.filename
        verbose = ''
        if Msg.level >= Msg.VER:
            verbose = 'v'
        cmd = ["tar", "-C", sourcedir, "-c" + verbose, "--one-file-system",
               "-S", "--xattrs", "-f", tarfile, "."]
        status = Uprocess().call(cmd, stderr=Msg.chlderr, close_fds=True)
        if status:
            Msg().err("Error: creating tar file:", tarfile)
        return not status

    def copydir(self, destdir, sourcedir=None):
        """Copy directories"""
        if sourcedir is None:
            sourcedir = self.filename
        verbose = ''
        if Msg.level >= Msg.VER:
            verbose = 'v'
        cmd_tarc = ["tar", "-C", sourcedir, "-c" + verbose,
                    "--one-file-system", "-S", "--xattrs", "-f", "-", "."]
        cmd_tarx = ["tar", "-C", destdir, "-x" + verbose, "-f", "-"]
        status = Uprocess().pipe(cmd_tarc, cmd_tarx)
        if not status:
            Msg().err("Error: copying:", sourcedir, " to ", destdir, l=Msg.VER)
        return status

    def cleanup(self):
        """Delete all temporary files"""
        tmptrash_copy = dict(FileUtil.tmptrash)
        for filename in tmptrash_copy:
            FileUtil(filename).remove(recursive=True)

    def isdir(self):
        """Is filename a directory"""
        try:
            if os.path.isdir(self.filename):
                return True
        except (IOError, OSError, TypeError):
            pass
        return False

    def size(self):
        """File size in bytes"""
        try:
            fstat = os.stat(self.filename)
            return fstat.st_size
        except (IOError, OSError, TypeError):
            return -1

    def getdata(self, mode="rb"):
        """Read file content to a buffer"""
        try:
            with open(self.filename, mode) as filep:
                buf = filep.read()
            Msg().out("Info: read buf", buf, l=Msg.DBG)
            return buf
        except (IOError, OSError, TypeError):
            return ""

    def get1stline(self, mode="rb"):
        """Read file 1st line to a buffer"""
        try:
            filep = open(self.filename, mode)
        except (IOError, OSError, TypeError):
            return ""
        else:
            buf = filep.readline().strip()
            filep.close()
            return buf

    def putdata(self, buf, mode="wb"):
        """Write buffer to file"""
        try:
            with open(self.filename, mode) as filep:
                filep.write(buf)
            return buf
        except (IOError, OSError, TypeError):
            return ""

    def getvalid_path(self):
        """Get the portion of a pathname that exists"""
        f_path = self.filename
        while f_path:
            if os.path.exists(f_path):
                return f_path
            (f_path, dummy) = os.path.split(f_path)
        return f_path

    def _cont2host(self, pathname, container_root, volumes=""):
        """Auxiliary translate container path to host path"""
        if not (pathname and pathname.startswith('/')):
            return ""
        if not volumes:
            volumes = []
        path = ""
        real_container_root = os.path.realpath(container_root)
        pathname = re.sub("/+", '/', os.path.normpath(pathname))
        for vol in volumes:
            (host_path, cont_path) = Uvolume(vol).split()
            if cont_path != host_path:
                if pathname.startswith(cont_path):
                    path = host_path + pathname[len(cont_path):]
                    break
            elif pathname.startswith(host_path):
                path = pathname
                break
        if not path:
            path = real_container_root + '/' + pathname
        f_path = ""
        for d_comp in path.split('/')[1:]:
            f_path = f_path + '/' + d_comp
            while os.path.islink(f_path):
                real_path = os.readlink(f_path)
                if real_path.startswith('/'):
                    if f_path.startswith(real_container_root): # in container
                        if real_path.startswith(real_container_root):
                            f_path = real_path
                        else:
                            f_path = real_container_root + real_path
                    else:
                        f_path = real_path
                else:
                    f_path = os.path.dirname(f_path) + '/' + real_path
        return os.path.realpath(f_path)

    def cont2host(self, container_path, volumes=""):
        """Translate container relative path to host path"""
        return self._cont2host(container_path, self.orig_filename, volumes)

    def _find_exec(self, path, rootdir="", volumes="", workdir="",
                   cont2host=False):
        """Find file in a path set such as PATH=/usr/bin:/bin"""
        for directory in path:
            if not directory:
                continue
            if directory == "." and workdir:
                directory = workdir
            elif directory == ".." and workdir:
                directory = workdir + "/.."
            if self.orig_filename.startswith("/"):
                exec_path = self.orig_filename
            else:
                exec_path = directory + "/" + self.orig_filename
            host_path = exec_path
            if rootdir:
                host_path = self._cont2host(exec_path, rootdir, volumes)
            if os.path.isfile(host_path) and os.access(host_path, os.X_OK):
                return host_path if cont2host else exec_path
        return ""

    def find_exec(self, path="", rootdir="", volumes="", workdir="",
                  cont2host=False):
        """Find an executable pathname"""
        if not path:
            path = os.getenv("PATH") + ":" + Config.conf['root_path']
        if rootdir:
            rootdir += "/"
        if is_genstr(path):
            if "=" in path:
                path = "".join(path.split("=", 1)[1:])
            path = path.split(":")
        if not isinstance(path, (list, tuple)):
            return ""
        return self._find_exec(path, rootdir, volumes, workdir, cont2host)

    def rename(self, dest_filename):
        """Rename/move file"""
        try:
            os.rename(self.filename, dest_filename)
        except (IOError, OSError):
            return False
        return True

    def _stream2file(self, dest_filename, mode="w"):
        """Copy from stdin to another file. We avoid shutil to have
        the fewest possible dependencies on other Python modules.
        """
        try:
            fpdst = open(dest_filename, mode + "b")
        except (IOError, OSError):
            return False
        while True:
            copy_buffer = sys.stdin.read(1024 * 1024)
            if not copy_buffer:
                break
            fpdst.write(copy_buffer)
        fpdst.close()
        return True

    def _file2stream(self):
        """Copy self.filename to stdout. We avoid shutil to have
        the fewest possible dependencies on other Python modules.
        """
        try:
            fpsrc = open(self.filename, "rb")
        except (IOError, OSError):
            return False

        while True:
            copy_buffer = fpsrc.read(1024 * 1024)
            if not copy_buffer:
                break
            sys.stdout.write(copy_buffer)

        fpsrc.close()
        return True

    def _file2file(self, dest_filename, mode="w"):
        """Copy self.filename to another file. We avoid shutil to have
        the fewest possible dependencies on other Python modules.
        """
        try:
            fpsrc = open(self.filename, "rb")
        except (IOError, OSError):
            return False
        try:
            fpdst = open(dest_filename, mode + "b")
        except (IOError, OSError):
            fpsrc.close()
            return False
        while True:
            copy_buffer = fpsrc.read(1024 * 1024)
            if not copy_buffer:
                break
            fpdst.write(copy_buffer)
        fpsrc.close()
        fpdst.close()
        return True

    def copyto(self, dest_filename, mode="w"):
        """Copy self.filename to another file. We avoid shutil to have
        the fewest possible dependencies on other Python modules.
        """
        if self.filename == "-" and dest_filename != "-":
            return self._stream2file(dest_filename, mode)

        if self.filename != "-" and dest_filename == "-":
            return self._file2stream()

        if self.filename != "-" and dest_filename != "-":
            return self._file2file(dest_filename, mode)

        return False

    def find_file_in_dir(self, image_list):
        """Find and return first file of list in dir"""
        path_prefix = self.filename
        for image in image_list:
            image_path = path_prefix + "/" + image
            if os.path.exists(image_path):
                return image_path
        return ""

    def _link_change_apply(self, new_l_path, f_path, force):
        """Actually apply the link convertion"""
        p_path = os.path.realpath(os.path.dirname(f_path))
        if force and not os.access(p_path, os.W_OK):
            os.chmod(p_path,
                     stat.S_IMODE(os.stat(p_path).st_mode) | stat.S_IWUSR)
            os.remove(f_path)
            os.symlink(new_l_path, f_path)
            os.chmod(p_path,
                     stat.S_IMODE(os.stat(p_path).st_mode) & ~stat.S_IWUSR)
        else:
            os.remove(f_path)
            os.symlink(new_l_path, f_path)

    def _link_set(self, f_path, orig_path, root_path, force):
        """Convertion to container specific symbolic link"""
        l_path = os.readlink(f_path)
        if not l_path.startswith("/"):
            return False
        new_l_path = ""
        regexp_id = "[a-z0-9]+-[a-z0-9]+-[a-z0-9]+-[a-z0-9]+-[a-z0-9]+"
        recomp = re.compile("(/.*/containers/" + regexp_id + "/ROOT)(/.*)")
        if orig_path == "":
            match = recomp.match(l_path)
            if match:
                orig_path = match.group(1)
        if (orig_path and l_path.startswith(orig_path) and
                orig_path != root_path):
            new_l_path = l_path.replace(orig_path, root_path, 1)
        elif not l_path.startswith(root_path):
            new_l_path = root_path + l_path
        if new_l_path:
            self._link_change_apply(new_l_path, f_path, force)
            return True
        return False

    def _link_restore(self, f_path, orig_path, root_path, force):
        """Convertion for host specific symbolic link"""
        l_path = os.readlink(f_path)
        new_l_path = ""
        if not l_path.startswith("/"):
            return False
        regexp_id = "[a-z0-9]+-[a-z0-9]+-[a-z0-9]+-[a-z0-9]+-[a-z0-9]+"
        recomp = re.compile("(/.*/containers/" + regexp_id + "/ROOT)(/.*)")
        if orig_path and l_path.startswith(orig_path):
            new_l_path = l_path.replace(orig_path, "", 1)
        elif l_path.startswith(root_path):
            new_l_path = l_path.replace(root_path, "", 1)
        elif orig_path == "":
            match = recomp.match(l_path)
            if match:
                new_l_path = l_path.replace(match.group(1), "", 1)
        if new_l_path:
            self._link_change_apply(new_l_path, f_path, force)
            return True
        return False

    def links_conv(self, force=False, to_container=True, orig_path=""):
        """ Convert absolute symbolic links to relative symbolic links"""
        root_path = os.path.realpath(self.filename)
        links = []
        if not self._is_safe_prefix(root_path):
            Msg().err("Error: links convertion outside of directory tree: ",
                      root_path)
            return None
        for dir_path, dirs, files in os.walk(root_path):
            for f_name in files + dirs:
                try:
                    f_path = dir_path + "/" + f_name
                    if not os.path.islink(f_path):
                        continue
                    if os.lstat(f_path).st_uid != HostInfo.uid:
                        continue
                    if to_container:
                        if self._link_set(f_path, orig_path, root_path, force):
                            links.append(f_path)
                    else:
                        if self._link_restore(f_path, orig_path,
                                              root_path, force):
                            links.append(f_path)
                except OSError:
                    continue
        return links

    def match(self):
        """Find matching file with wildcard matching expression"""
        directory = os.path.dirname(self.filename)
        matching_expression = os.path.basename(self.filename)
        matching_files = []
        if not os.path.isdir(directory):
            return []
        for f_name in os.listdir(directory):
            if re.match(matching_expression, f_name):
                matching_files.append(directory + "/" + f_name)
        return matching_files



# -*- coding: utf-8 -*-
"""Docker API integration"""

import os

from udocker.msg import Msg
from udocker.commonlocalfile import CommonLocalFileApi
from udocker.utils.fileutil import FileUtil
from udocker.helper.unique import Unique


class OciLocalFileAPI(CommonLocalFileApi):
    """Manipulate OCI container and/or image files"""

    def __init__(self, localrepo):
        CommonLocalFileApi.__init__(self, localrepo)

    def _load_structure(self, tmp_imagedir):
        """Load OCI image structure"""
        structure = {}
        structure["repolayers"] = {}
        structure["manifest"] = {}
        f_path = tmp_imagedir + '/'
        structure["oci-layout"] = \
                self.localrepo.load_json(f_path + "oci-layout")
        structure["index"] = \
                self.localrepo.load_json(f_path + "index.json")
        if not (structure["index"] and structure["oci-layout"]):
            return {}
        for fname in os.listdir(tmp_imagedir + "/blobs"):
            f_path = tmp_imagedir + "/blobs/" + fname
            if FileUtil(f_path).isdir():
                layer_algorithm = fname
                for layer_f in os.listdir(f_path):
                    layer_id = layer_algorithm + ':' + layer_f
                    structure["repolayers"][layer_id] = {}
                    structure["repolayers"][layer_id]["layer_f"] = \
                            f_path + '/' + layer_f
                    structure["repolayers"][layer_id]["layer_a"] = \
                            layer_algorithm
                    structure["repolayers"][layer_id]["layer_h"] = \
                            layer_f
        return structure

    def _get_from_manifest(self, structure, imagetag):
        """Search for OCI manifest and return Config and layer ids"""
        try:
            config_layer = \
                    structure["manifest"][imagetag]["json"]["config"]["digest"]
        except (KeyError, ValueError, TypeError):
            config_layer = ""
        try:
            layers = []
            for layer in structure["manifest"][imagetag]["json"]["layers"]:
                layers.append(layer["digest"])
            layers.reverse()
            return (config_layer, layers)
        except (KeyError, ValueError, TypeError):
            return (config_layer, [])

    def _load_manifest(self, structure, manifest):
        """Load OCI manifest"""
        try:
            tag = manifest["annotations"]["org.opencontainers.image.ref.name"]
        except KeyError:
            tag = Unique().imagetag()
        if '/' in tag and ':' in tag:
            (imagerepo, tag) = tag.split(':', 1)
        else:
            imagerepo = Unique().imagename()
        if self._imagerepo:
            imagerepo = self._imagerepo
        imagetag = imagerepo + ':' + tag
        structure["manifest"][imagetag] = {}
        structure["manifest"][imagetag]["json"] = \
                self.localrepo.load_json(structure["repolayers"]\
                [manifest["digest"]]["layer_f"])
        structure["manifest"][imagetag]["json_f"] = \
                structure["repolayers"][manifest["digest"]]["layer_f"]
        return self._load_image(structure, imagerepo, tag)

    def _load_repositories(self, structure):
        """Load OCI image repositories"""
        loaded_repositories = []
        for manifest in structure["index"]["manifests"]:
            if manifest["mediaType"] == \
                    "application/vnd.oci.image.manifest.v1+json":
                loaded_repositories.append(
                    self._load_manifest(structure, manifest))
            elif manifest["mediaType"] == \
                    "application/vnd.oci.image.index.v1+json":
                loaded_repositories.extend(self._load_repositories(
                    self.localrepo.load_json(
                        structure["repolayers"]
                        [manifest["digest"]]["layer_f"])))
        return loaded_repositories

    def _load_image_step2(self, structure, imagerepo, tag):
        """Prepare load of OCI image"""
        imagetag = imagerepo + ':' + tag
        (config_layer_id, layers) = \
                self._get_from_manifest(structure, imagetag)
        if config_layer_id:
            json_file = structure["repolayers"][config_layer_id]["layer_f"]
            self._move_layer_to_v1repo(json_file, config_layer_id,
                                       "container.json")
        layer_hash_list = []
        for layer_id in layers:
            Msg().out("Info: adding layer:", layer_id, l=Msg.INF)
            filename = str(structure["repolayers"][layer_id]["layer_f"])
            if not self._move_layer_to_v1repo(filename, layer_id):
                Msg().err("Error: copying layer file", filename, l=Msg.VER)
                return []
            layer_hash_list.append(structure["repolayers"][layer_id]["layer_h"])
        self.localrepo.save_json("ancestry", layer_hash_list)
        return [imagetag]

    def load(self, tmp_imagedir, imagerepo=None):
        """Load an OCI image file The file is a tarball containing several
        layers, each layer has metadata and data content (directory tree)
        stored as a tar file.
        """
        self._imagerepo = imagerepo
        structure = self._load_structure(tmp_imagedir)
        if not structure:
            Msg().err("Error: failed to load image structure")
            return []
        return self._load_repositories(structure)



#!/usr/bin/env python
"""
udocker unit tests: OSInfo
"""

from unittest import TestCase, main
from unittest.mock import patch
from udocker.helper.osinfo import OSInfo
from udocker.config import Config


class GuestInfoTestCase(TestCase):
    """Test OSInfo() class."""

    def setUp(self):
        Config().getconf()
        self.rootdir = "~/.udocker/container/abcd0/ROOT"
        self.file = "/bin/ls"
        self.noosdist = ("", "xx")

    def tearDown(self):
        pass

    def test_01_init(self):
        """Test01 OSInfo() constructor."""
        ginfo = OSInfo(self.rootdir)
        self.assertIsInstance(ginfo._binarylist, list)

    @patch('udocker.helper.osinfo.Uprocess.get_output')
    @patch('udocker.helper.osinfo.os.path.isfile')
    def test_02_get_filetype(self, mock_isfile, mock_getout):
        """Test02 OSInfo.get_filetype(filename)"""
        # full filepath exists
        ftype = "/bin/ls: yyy, x86-64, xxx"
        mock_isfile.return_value = True
        mock_getout.return_value = ftype
        ginfo = OSInfo(self.rootdir)
        status = ginfo.get_filetype(self.file)
        self.assertEqual(status, ftype)

        # file does not exist
        nofile = "ddd: cannot open"
        mock_isfile.return_value = False
        mock_getout.return_value = nofile
        ginfo = OSInfo(self.rootdir)
        status = ginfo.get_filetype(nofile)
        self.assertEqual(status, "")

    @patch.object(OSInfo, 'get_filetype')
    def test_03_arch(self, mock_getftype):
        """Test03 OSInfo.arch()"""
        # arch is x86_64
        ftype = "/bin/ls: yyy, x86-64, xxx"
        mock_getftype.return_value = ftype
        ginfo = OSInfo(self.rootdir)
        status = ginfo.arch()
        self.assertEqual(status, "amd64")

        # arch is i386
        ftype = "/bin/ls: yyy, Intel 80386, xxx"
        mock_getftype.return_value = ftype
        ginfo = OSInfo(self.rootdir)
        status = ginfo.arch()
        self.assertEqual(status, "i386")

        # arch is arm 64
        ftype = "/bin/ls: yyy, aarch64"
        mock_getftype.return_value = ftype
        ginfo = OSInfo(self.rootdir)
        status = ginfo.arch()
        self.assertEqual(status, "arm64")

        # arch is arm
        ftype = "/bin/ls: yyy, ARM, xxx"
        mock_getftype.return_value = ftype
        ginfo = OSInfo(self.rootdir)
        status = ginfo.arch()
        self.assertEqual(status, "arm")

    @patch('udocker.helper.osinfo.os.path.exists')
    @patch('udocker.helper.osinfo.FileUtil.match')
    @patch('udocker.helper.osinfo.FileUtil.getdata')
    def test_04_osdistribution(self, mock_gdata, mock_match, mock_exists):
        """Test04 OSInfo.osdistribution()"""
        lsbdata = "DISTRIB_ID=Ubuntu\n" \
                  "DISTRIB_RELEASE=16.04\n" \
                  "DISTRIB_CODENAME=xenial\n" \
                  "DISTRIB_DESCRIPTION=Ubuntu 16.04.5 LTS\n"
        mock_match.return_value = ["/etc/lsb-release"]
        osdist = ("Ubuntu", "16")
        mock_exists.return_value = True
        mock_gdata.return_value = lsbdata
        ginfo = OSInfo(self.rootdir)
        status = ginfo.osdistribution()
        self.assertEqual(status, osdist)

        reldata = "CentOS release 6.10 (Final)"
        mock_match.return_value = ["/etc/centos-release"]
        osdist = ("CentOS", "6")
        mock_exists.return_value = True
        mock_gdata.return_value = reldata
        ginfo = OSInfo(self.rootdir)
        status = ginfo.osdistribution()
        self.assertEqual(status, osdist)

    @patch.object(OSInfo, 'osdistribution')
    def test_05_osversion(self, mock_osdist):
        """Test05 OSInfo.osversion()"""
        # has osdistro
        osdist = ("Ubuntu", "16.04")
        mock_osdist.return_value = osdist
        ginfo = OSInfo(self.rootdir)
        status = ginfo.osversion()
        self.assertEqual(status, "linux")

        # has no osdistro
        mock_osdist.return_value = self.noosdist
        ginfo = OSInfo(self.rootdir)
        status = ginfo.osversion()
        self.assertEqual(status, "")


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Configurations options and treatment/overrinding"""
import os
import sys
from udocker.msg import Msg

# if Python 3
if sys.version_info[0] >= 3:
    from configparser import ConfigParser
else:
    from ConfigParser import ConfigParser


class Config(object):
    """Default configuration values for the whole application. Changes
    to these values should be made via a configuration file read via
    Config.init() and that can reside in ~/.udocker/udocker.conf
    """
    conf = {}
    conf['verbose_level'] = 3
    conf['homedir'] = os.path.expanduser("~") + "/.udocker"
    conf['topdir'] = conf['homedir']
    conf['bindir'] = None
    conf['libdir'] = None
    conf['docdir'] = None
    conf['reposdir'] = None
    conf['layersdir'] = None
    conf['containersdir'] = None

    # udocker installation tarball the release is the minimum requirement
    # the actual tarball used in the installation can have a higher version
    conf['tarball_release'] = "1.2.8"
    conf['tarball'] = (
        "https://download.ncg.ingrid.pt/"
        "webdav/udocker/udocker-englib-1.2.8.tar.gz"
        " "
        "https://raw.githubusercontent.com"
        "/jorge-lip/udocker-builds/master/tarballs/udocker-englib-1.2.8.tar.gz"
    )
    conf['installinfo'] = [
        "https://raw.githubusercontent.com/indigo-dc/udocker/master/messages", ]
    conf['installretry'] = 3
    conf['autoinstall'] = True
    conf['config'] = "udocker.conf"
    conf['keystore'] = "keystore"
    conf['tmpdir'] = os.getenv("TMPDIR", "/tmp")    # for tmp files only

    # defaults for container execution
    conf['cmd'] = ["/bin/bash", "-i"]  # Comand to execute

    # default path for executables
    conf['root_path'] = "/usr/sbin:/sbin:/usr/bin:/bin"
    conf['user_path'] = "/usr/local/bin:/usr/bin:/bin"

    # directories to be mapped in containers with: run --sysdirs
    conf['sysdirs_list'] = ("/dev", "/proc", "/sys",
                            "/etc/resolv.conf", "/etc/host.conf",
                            "/lib/modules", )

    # POSSIBLE DEPRECATED
    # directories to be mapped in containers with: run --hostauth
    # conf['hostauth_list'] = ("/etc/passwd", "/etc/group",
    #                               "/etc/shadow", "/etc/gshadow", )

    # directories for DRI (direct rendering)
    conf['dri_list'] = ("/usr/lib64/dri", "/lib64/dri",
                        "/usr/lib/dri", "/lib/dri", )

    # allowed file mountpoints for runC, these files can be copied in
    conf['mountpoint_prefixes'] = ("/etc", )

    # container execution mode if not set via setup
    # Change it to P2 if execution problems occur
    conf['default_execution_mode'] = "P1"

    # PRoot override seccomp
    # conf['proot_noseccomp'] = True
    conf['proot_noseccomp'] = None
    conf['proot_killonexit'] = True   # PRoot kill-on-exit

    # fakechroot engine get ld_library_paths from ld.so.cache
    conf['ld_so_cache'] = "/etc/ld.so.cache"

    # fakechroot engine override fakechroot.so selection
    # conf['fakechroot_so'] = "libfakechroot-CentOS-7-x86_64.so"
    conf['fakechroot_so'] = None

    # translate symbolic links in pathnames None means automatic
    conf['fakechroot_expand_symlinks'] = None

    # sharable library directories
    conf['lib_dirs_list_x86_64'] = ("/usr/lib/x86_64-linux-gnu",
                                    "/usr/lib64",)

    # fakechroot sharable library directories
    conf['lib_dirs_list_essential'] = ("/lib/x86_64-linux-gnu",
                                       "/usr/lib/x86_64-linux-gnu",
                                       "/lib64", "/usr/lib64", "/lib",
                                       "/usr/lib", )
    conf['lib_dirs_list_append'] = (".", )

    # fakechroot access files, used to circunvent openmpi init issues
    conf['access_files'] = ("/sys/class/infiniband", "/dev/open-mx",
                            "/dev/myri0", "/dev/myri1", "/dev/myri2",
                            "/dev/myri3", "/dev/myri4", "/dev/myri5",
                            "/dev/myri6", "/dev/myri7", "/dev/myri8",
                            "/dev/myri9", "/dev/ipath", "/dev/kgni0",
                            "/dev/mic/scif", "/dev/scif", )

    # Force the use of specific executables
    # UDOCKER = use executable from the udocker binary distribution/tarball
    conf['use_proot_executable'] = "UDOCKER"
    conf['use_runc_executable'] = ""
    conf['use_singularity_executable'] = ""

    # runc parameters
    conf['runc_nomqueue'] = None   # runc parameters
    conf['runc_capabilities'] = [
        "CAP_KILL", "CAP_NET_BIND_SERVICE", "CAP_CHOWN", "CAP_DAC_OVERRIDE",
        "CAP_FOWNER", "CAP_FSETID", "CAP_KILL", "CAP_SETGID", "CAP_SETUID",
        "CAP_SETPCAP", "CAP_NET_BIND_SERVICE", "CAP_NET_RAW",
        "CAP_SYS_CHROOT", "CAP_MKNOD", "CAP_AUDIT_WRITE", "CAP_SETFCAP",
    ]

    # singul. opts -u --nv -w
    conf['singularity_options'] = ["-w", ]

    # Pass host env vars
    conf['valid_host_env'] = ("TERM", "PATH", "PROOT_TMP_DIR", )
    conf['invalid_host_env'] = ("VTE_VERSION", )

    # CPU affinity executables to use with: run --cpuset-cpus="1,2,3-4"
    conf['cpu_affinity_exec_tools'] = (["numactl", "-C", "%s", "--", ],
                                       ["taskset", "-c", "%s", ])

    # Containers execution defaults
    conf['location'] = ""      # run container in this location

    # Curl settings
    conf['http_proxy'] = ""    # ex. socks5://user:pass@127.0.0.1:1080
    conf['timeout'] = 12       # default timeout (secs)
    conf['download_timeout'] = 30 * 60  # file download timeout (secs)
    conf['ctimeout'] = 6       # default TCP connect timeout (secs)
    conf['http_agent'] = ""
    conf['http_insecure'] = False
    conf['use_curl_executable'] = ""  # force use of executable

    # docker hub index
    conf['dockerio_index_url'] = "https://hub.docker.com"
    # docker hub registry
    #conf['dockerio_registry_url'] = "https://registry.hub.docker.com"
    conf['dockerio_registry_url'] = "https://registry-1.docker.io"
    # private repository v2
    # conf['dockerio_registry_url'] = "http://localhost:5000"

    # registries table
    conf['docker_registries'] = {"docker.io": [conf['dockerio_registry_url'],
                                               conf['dockerio_index_url']],
                                }

    # nvidia files
    conf['nvi_etc_list'] = ['vulkan/icd.d/nvidia_icd.json',
                            'OpenCL/vendors/nvidia.icd']

    conf['nvi_bin_list'] = ['nvidia-bug-report.sh',
                            'nvidia-cuda-mps-control',
                            'nvidia-cuda-mps-server',
                            'nvidia-debugdump', 'nvidia-installer',
                            'nvidia-persistenced', 'nvidia-settings',
                            'nvidia-smi', 'nvidia-uninstall',
                            'nvidia-xconfig']

    conf['nvi_lib_list'] = ['libOpenCL.', 'libcuda.', 'libnvcuvid.',
                            'libnvidia-cfg.', 'libnvidia-compiler.',
                            'libnvidia-encode.',
                            'libnvidia-fatbinaryloader.',
                            'libnvidia-fbc.', 'libnvidia-ifr.',
                            'libnvidia-ml.', 'libnvidia-opencl.',
                            'libnvidia-ptxjitcompiler.',
                            'libnvidia-tls.', 'tls/libnvidia-tls.']

    conf['nvi_dev_list'] = ['/dev/nvidia', ]

    def _conf_file_read(self, cfpath, ignore_keys=None):
        """
        Read config file
        """
        Msg().out('Info: using config file: ', cfpath)
        cfnparser = ConfigParser()
        cfnparser.read(cfpath)
        for (key, val) in cfnparser.items('DEFAULT'):
            if ignore_keys and key in ignore_keys:
                continue
            if val is not None:
                Config.conf[key] = val

    def _file_override(self, user_cfile, ignore_keys=None):
        """
        Override values from config file
        """
        if os.path.exists('/etc/' + Config.conf['config']):
            self._conf_file_read('/etc/' + Config.conf['config'], ignore_keys)

        cfpath = Config.conf['homedir'] + '/' + Config.conf['config']
        if os.path.exists(cfpath):
            self._conf_file_read(cfpath, ignore_keys)

        if Config.conf['topdir'] != Config.conf['homedir']:
            cfpath = Config.conf['topdir'] + '/'  + Config.conf['config']
            if os.path.exists(cfpath):
                self._conf_file_read(cfpath, ignore_keys)

        if os.path.exists(user_cfile):
            self._conf_file_read(user_cfile, ignore_keys)

    def _env_override(self):
        """Override config with environment"""
        Config.conf['verbose_level'] = \
            int(os.getenv("UDOCKER_LOGLEVEL", Config.conf['verbose_level']))
        Config.conf['topdir'] = os.getenv("UDOCKER_DIR", Config.conf['topdir'])
        Config.conf['bindir'] = os.getenv("UDOCKER_BIN", Config.conf['bindir'])
        Config.conf['libdir'] = os.getenv("UDOCKER_LIB", Config.conf['libdir'])
        Config.conf['docdir'] = os.getenv("UDOCKER_DOC", Config.conf['docdir'])
        Config.conf['reposdir'] = \
            os.getenv("UDOCKER_REPOS", Config.conf['reposdir'])
        Config.conf['layersdir'] = \
            os.getenv("UDOCKER_LAYERS", Config.conf['layersdir'])
        Config.conf['containersdir'] = \
            os.getenv("UDOCKER_CONTAINERS", Config.conf['containersdir'])
        Config.conf['dockerio_index_url'] = \
            os.getenv("UDOCKER_INDEX", Config.conf['dockerio_index_url'])
        Config.conf['dockerio_registry_url'] = \
            os.getenv("UDOCKER_REGISTRY", Config.conf['dockerio_registry_url'])
        Config.conf['tarball'] = \
            os.getenv("UDOCKER_TARBALL", Config.conf['tarball'])
        Config.conf['default_execution_mode'] = \
            os.getenv("UDOCKER_DEFAULT_EXECUTION_MODE",
                      Config.conf['default_execution_mode'])
        Config.conf['fakechroot_so'] = \
            os.getenv("UDOCKER_FAKECHROOT_SO", Config.conf['fakechroot_so'])
        Config.conf['tmpdir'] = os.getenv("UDOCKER_TMP", Config.conf['tmpdir'])
        Config.conf['keystore'] = \
            os.getenv("UDOCKER_KEYSTORE", Config.conf['keystore'])
        Config.conf['use_curl_executable'] = \
            os.getenv("UDOCKER_USE_CURL_EXECUTABLE",
                      Config.conf['use_curl_executable'])
        Config.conf['use_proot_executable'] = \
            os.getenv("UDOCKER_USE_PROOT_EXECUTABLE",
                      Config.conf['use_proot_executable'])
        Config.conf['use_runc_executable'] = \
            os.getenv("UDOCKER_USE_RUNC_EXECUTABLE",
                      Config.conf['use_runc_executable'])
        Config.conf['use_singularity_executable'] = \
            os.getenv("UDOCKER_USE_SINGULARITY_EXECUTABLE",
                      Config.conf['use_singularity_executable'])

        Config.conf['fakechroot_expand_symlinks'] = \
            os.getenv("UDOCKER_FAKECHROOT_EXPAND_SYMLINKS",
                      str(Config.conf['fakechroot_expand_symlinks'])).lower()
        os.environ["PROOT_TMP_DIR"] = os.getenv("PROOT_TMP_DIR",
                                                Config.conf['tmpdir'])

    def getconf(self, user_cfile="u.conf"):
        """Return all configuration variables"""
        self._file_override(user_cfile) # Override with variables in conf file
        self._env_override()          # Override with variables in environment

    def container(self, user_cfile="u.conf"):
        """
        Load configuration for a container
        Values should be in the form x = y
        """
        ignore_keys = ["topdir", "homedir", "reposdir", "layersdir",
                       "containersdir", "location", ]
        self._file_override(user_cfile, ignore_keys)



#!/usr/bin/env python
"""
udocker unit tests: UMain
"""

from unittest import TestCase, main
from unittest.mock import patch
from udocker.umain import UMain
from udocker.config import Config


class UMainTestCase(TestCase):
    """Test UMain() class main udocker program."""

    def setUp(self):
        Config().getconf()

    def tearDown(self):
        pass

    def test_01_init(self):
        """Test01 UMain(argv) constructor."""
        argv = ["udocker", "-h"]
        udoc = UMain(argv)
        self.assertEqual(udoc.argv, argv)

    @patch('udocker.umain.Msg')
    @patch('udocker.umain.UdockerCLI')
    @patch('udocker.umain.LocalRepository')
    @patch('udocker.umain.os.geteuid')
    def test_02__prepare_exec(self, mock_getuid,
                              mock_local, mock_ucli, mock_msg):
        """Test02 UMain()._prepare_exec()."""
        argv = ["udocker", "-h"]
        mock_msg.level = 0
        mock_msg.VER = 4
        mock_getuid.return_value = 0
        with patch('sys.exit') as mock_exit:
            umain = UMain(argv)
            umain._prepare_exec()
            self.assertTrue(mock_exit.called)

        argv = ["udocker", "-h", "--debug", "--insecure"]
        mock_msg.level = 0
        mock_msg.VER = 4
        mock_getuid.return_value = 100
        mock_local.return_value.is_repo.return_value = True
        mock_local.return_value.create_repo.return_value = None
        mock_ucli.return_value = None
        umain = UMain(argv)
        umain._prepare_exec()
        self.assertTrue(mock_getuid.called)
        self.assertTrue(mock_local.return_value.is_repo.called)
        self.assertTrue(mock_ucli.called)

    @patch('udocker.umain.Msg')
    @patch('udocker.umain.UdockerCLI')
    def test_03_execute(self, mock_ucli, mock_msg):
        """Test03 UMain().execute()."""
        mock_msg.level = 0
        argv = ['udocker', '--allow-root', '-h']
        mock_ucli.return_value.do_help.return_value = 0
        umain = UMain(argv)
        status = umain.execute()
        self.assertTrue(mock_ucli.return_value.do_help.called)
        self.assertEqual(status, 0)

        argv = ['udocker', '--allow-root', '--version']
        mock_ucli.return_value.do_version.return_value = 0
        umain = UMain(argv)
        status = umain.execute()
        self.assertTrue(mock_ucli.return_value.do_version.called)
        self.assertEqual(status, 0)

        argv = ['udocker', '--allow-root', 'install']
        mock_ucli.return_value.do_install.return_value = 0
        umain = UMain(argv)
        status = umain.execute()
        self.assertTrue(mock_ucli.return_value.do_install.called)
        self.assertEqual(status, 0)

        argv = ['udocker', '--allow-root', 'showconf']
        mock_ucli.return_value.do_showconf.return_value = 0
        umain = UMain(argv)
        status = umain.execute()
        self.assertTrue(mock_ucli.return_value.do_showconf.called)
        self.assertEqual(status, 0)

        argv = ['udocker', '--allow-root', 'rm']
        mock_ucli.return_value.do_rm.return_value = 0
        umain = UMain(argv)
        status = umain.execute()
        self.assertTrue(mock_ucli.return_value.do_rm.called)
        self.assertEqual(status, 0)

        argv = ['udocker', '--allow-root', 'faking']
        umain = UMain(argv)
        status = umain.execute()
        self.assertEqual(status, 1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

# -*- coding: utf-8 -*-
"""
udocker unit tests: CurlHeader
"""

from unittest import TestCase, main
from unittest.mock import patch
from io import StringIO
from udocker.utils.curl import CurlHeader
from udocker.utils.curl import GetURL
from udocker.utils.curl import GetURLpyCurl
from udocker.utils.curl import GetURLexeCurl
from udocker.config import Config

BUILTINS = "builtins"


class CurlHeaderTestCase(TestCase):
    """Test CurlHeader() http header parser."""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_01_init(self):
        """Test01 CurlHeader() constructor."""
        curl_header = CurlHeader()
        self.assertFalse(curl_header.sizeonly)
        self.assertIsInstance(curl_header.data, dict)
        self.assertEqual("", curl_header.data["X-ND-HTTPSTATUS"])
        self.assertEqual("", curl_header.data["X-ND-CURLSTATUS"])

    def test_02_write(self):
        """Test02 CurlHeader().write()."""
        buff = ["HTTP/1.1 200 OK",
                "Content-Type: application/octet-stream",
                "Content-Length: 32", ]
        curl_header = CurlHeader()
        for line in buff:
            curl_header.write(line)
        status = curl_header.data["content-type"]
        self.assertEqual(status, "application/octet-stream")
        status = curl_header.data["X-ND-HTTPSTATUS"]
        self.assertEqual(status, "HTTP/1.1 200 OK")

        curl_header = CurlHeader()
        for line in buff:
            curl_header.write(line)
        buff_out = curl_header.getvalue()
        self.assertTrue("HTTP/1.1 200 OK" in buff_out)

        line = ""
        curl_header = CurlHeader()
        curl_header.sizeonly = True
        self.assertEqual(-1, curl_header.write(line))

    @patch.object(CurlHeader, 'write')
    def test_03_setvalue_from_file(self, mock_write):
        """Test03 CurlHeader().setvalue_from_file()."""
        fakedata = StringIO('XXXX')
        with patch(BUILTINS + '.open') as mopen:
            mopen.return_value.__iter__ = (
                lambda self: iter(fakedata.readline, ''))
            curl_header = CurlHeader()
            self.assertTrue(curl_header.setvalue_from_file("filename"))
            self.assertTrue(mock_write.called_with('XXXX'))

    def test_04_getvalue(self):
        """Test04 CurlHeader().getvalue()."""
        curl_header = CurlHeader()
        curl_header.data = "XXXX"
        status = curl_header.getvalue()
        self.assertEqual(status, curl_header.data)


class GetURLTestCase(TestCase):
    """Test GetURL() perform http operations portably."""

    def setUp(self):
        Config().getconf()
        Config().conf['timeout'] = 1
        Config().conf['ctimeout'] = 1
        Config().conf['download_timeout'] = 1
        Config().conf['http_agent'] = ""
        Config().conf['http_proxy'] = ""
        Config().conf['http_insecure'] = 0
        Config().conf['use_curl_exec'] = ""

    def tearDown(self):
        pass

    def _get(self, *args, **kwargs):
        """Mock for pycurl.get."""
        return args[0]

    @patch.object(GetURLexeCurl, '_select_implementation')
    @patch.object(GetURLexeCurl, 'is_available')
    @patch.object(GetURLpyCurl, 'is_available')
    @patch('udocker.utils.curl.Msg')
    def test_01_init(self, mock_msg, mock_gupycurl,
                     mock_guexecurl, mock_select):
        """Test01 GetURL() constructor."""
        mock_msg.level = 0
        mock_gupycurl.return_value = False
        mock_guexecurl.return_value = True
        geturl = GetURL()
        mock_select.assert_called()
        self.assertEqual(geturl.ctimeout, Config().conf['ctimeout'])
        self.assertEqual(geturl.insecure, Config().conf['http_insecure'])
        self.assertFalse(geturl.cache_support)

    @patch('udocker.utils.curl.Msg')
    @patch.object(GetURLexeCurl, 'is_available')
    @patch.object(GetURLpyCurl, 'is_available')
    def test_02__select_implementation(self, mock_gupycurl,
                                       mock_guexecurl, mock_msg):
        """Test02 GetURL()._select_implementation()."""
        mock_msg.level = 0
        mock_gupycurl.return_value = True
        geturl = GetURL()
        geturl._select_implementation()
        # self.assertTrue(geturl.cache_support)

        mock_gupycurl.return_value = False
        geturl = GetURL()
        geturl._select_implementation()
        self.assertFalse(geturl.cache_support)

        mock_guexecurl.return_value = False
        with self.assertRaises(NameError):
            GetURL()

    def test_03_get_content_length(self):
        """Test03 GetURL().get_content_length()."""
        hdr = type('test', (object,), {})()
        hdr.data = {"content-length": 10, }
        geturl = GetURL()
        self.assertEqual(geturl.get_content_length(hdr), 10)

        hdr = type('test', (object,), {})()
        hdr.data = {"content-length": dict(), }
        geturl = GetURL()
        self.assertEqual(geturl.get_content_length(hdr), -1)

    @patch.object(GetURLpyCurl, 'is_available')
    def test_04_set_insecure(self, mock_gupycurl):
        """Test04 GetURL().set_insecure()."""
        mock_gupycurl.return_value = True
        geturl = GetURL()
        geturl.set_insecure()
        self.assertEqual(geturl.insecure, True)

        geturl = GetURL()
        geturl.set_insecure(False)
        self.assertEqual(geturl.insecure, False)

    @patch.object(GetURLpyCurl, 'is_available')
    def test_05_set_proxy(self, mock_gupycurl):
        """Test05 GetURL().set_proxy()."""
        mock_gupycurl.return_value = True
        geturl = GetURL()
        geturl.set_proxy("http://host")
        self.assertEqual(geturl.http_proxy, "http://host")

    def test_06_get(self):
        """Test06 GetURL().get()."""
        geturl = GetURL()
        self.assertRaises(TypeError, geturl.get)

        geturl = GetURL()
        geturl._geturl = type('test', (object,), {})()
        geturl._geturl.get = self._get
        self.assertEqual(geturl.get("http://host"), "http://host")

    def test_07_post(self):
        """Test07 GetURL().post()."""
        geturl = GetURL()
        self.assertRaises(TypeError, geturl.post)
        self.assertRaises(TypeError, geturl.post, "http://host")

        geturl = GetURL()
        geturl._geturl = type('test', (object,), {})()
        geturl._geturl.get = self._get
        status = geturl.post("http://host", {"DATA": 1, })
        self.assertEqual(status, "http://host")

    # def test_08_get_status_code(self):
    #     """Test08 GetURL().get_status_code()."""


class GetURLpyCurlTestCase(TestCase):
    """GetURLpyCurl TestCase."""

    def setUp(self):
        Config().getconf()
        Config().conf['timeout'] = 1
        Config().conf['ctimeout'] = 1
        Config().conf['download_timeout'] = 1
        Config().conf['http_agent'] = ""
        Config().conf['http_proxy'] = ""
        Config().conf['http_insecure'] = 0

    def tearDown(self):
        pass

    def _get(self, *args, **kwargs):
        """Mock for pycurl.get."""
        return args[0]

    # def test_01_init(self):
    #     """Test01 GetURLpyCurl() constructor."""

    @patch('udocker.utils.curl.Msg')
    @patch.object(GetURLpyCurl, 'is_available')
    def test_02_is_available(self, mock_gupycurl, mock_msg):
        """Test02 GetURLpyCurl()._is_available()."""
        mock_msg.level = 0
        mock_gupycurl.return_value = True
        geturl = GetURLpyCurl()
        geturl.is_available()
        self.assertTrue(geturl.is_available())

        mock_gupycurl.return_value = False
        geturl = GetURLpyCurl()
        self.assertFalse(geturl.is_available())

    # def test_03__select_implementation(self):
    #     """Test03 GetURLpyCurl()._select_implementation()."""

    @patch.object(GetURLpyCurl, 'is_available')
    @patch('udocker.utils.curl.Msg')
    @patch('udocker.utils.curl.pycurl')
    @patch('udocker.utils.curl.CurlHeader')
    def test_04__set_defaults(self, mock_hdr, mock_pyc,
                              mock_msg, mock_selinsec):
        """Test04 GetURLpyCurl()._set_defaults()."""
        mock_selinsec.return_value = True
        mock_msg.level = 0
        mock_msg.VER = 4
        geturl = GetURLpyCurl()
        geturl._set_defaults(mock_pyc, mock_hdr)
        self.assertTrue(mock_pyc.setopt.called)

        # when Msg.level >= Msg.VER = 4: AND insecure
        mock_msg.level = 5
        mock_msg.VER = 4
        geturl = GetURLpyCurl()
        geturl._set_defaults(mock_pyc, mock_hdr)
        self.assertEqual(mock_pyc.setopt.call_count, 18)

        mock_selinsec.return_value = True
        # when Msg.level < Msg.VER = 4: AND secure
        mock_msg.level = 2
        mock_msg.VER = 4
        geturl = GetURLpyCurl()
        geturl._set_defaults(mock_pyc, mock_hdr)
        self.assertEqual(mock_pyc.setopt.call_count, 27)

    # @patch.object(GetURLpyCurl, 'is_available')
    # @patch('udocker.utils.curl.Msg')
    # @patch('udocker.utils.curl.pycurl')
    # @patch('udocker.utils.curl.CurlHeader')
    # def test_05__mkpycurl(self, mock_hdr, mock_pyc, mock_msg, mock_sel):
    #     """Test05 GetURLpyCurl()._mkpycurl()."""
    #     mock_sel.return_value = True
    #     geturl = GetURLpyCurl()
    #     geturl._mkpycurl(mock_pyc, mock_hdr)
    #     self.assertTrue(mock_pyc.setopt.called)

    def test_06_get(self):
        """Test06 GetURLpyCurl().get() generic get."""
        geturl = GetURLpyCurl()
        geturl._geturl = type('test', (object,), {})()
        geturl.get = self._get
        status = geturl.get("http://host")
        self.assertEqual(status, "http://host")


class GetURLexeCurlTestCase(TestCase):
    """GetURLexeCurl TestCase."""

    def setUp(self):
        Config().getconf()
        Config().conf['timeout'] = 1
        Config().conf['ctimeout'] = 1
        Config().conf['download_timeout'] = 1
        Config().conf['http_agent'] = ""
        Config().conf['http_proxy'] = ""
        Config().conf['http_insecure'] = 0

    def tearDown(self):
        pass

    def _get(self, *args, **kwargs):
        """Mock for pycurl.get."""
        return args[0]

    def test_01_init(self):
        """Test01 GetURLexeCurl() constructor."""
        geturl = GetURLexeCurl()
        self.assertIsNone(geturl._opts)

        geturl = GetURLexeCurl()
        self.assertIsNone(geturl._files)

    @patch('udocker.utils.curl.Msg')
    @patch('udocker.utils.curl.FileUtil.find_exec')
    def test_02_is_available(self, mock_findexec, mock_msg):
        """Test02 GetURLexeCurl()._is_available()."""
        mock_msg.level = 0
        mock_findexec.return_value = "/tmp"
        geturl = GetURLexeCurl()
        self.assertTrue(geturl.is_available())

        mock_findexec.return_value = ""
        geturl = GetURLexeCurl()
        self.assertFalse(geturl.is_available())

    # def test_03__select_implementation(self):
    #     """Test03 GetURLexeCurl()._select_implementation()."""

    # @patch('udocker.utils.curl.Msg')
    # def test_04__set_defaults(self, mock_msg):
    #     """Test04 GetURLexeCurl()._set_defaults()"""
    #     mock_msg.return_value.level = 0
    #     mock_msg.VER = 4
    #     geturl = GetURLexeCurl()
    #     geturl._set_defaults()
    #     self.assertEqual(geturl._opts["insecure"], "")

    #     mock_msg.return_value.level = 0
    #     mock_msg.VER = 4
    #     geturl = GetURLexeCurl()
    #     geturl.insecure = True
    #     geturl._set_defaults()
    #     self.assertEqual(geturl._opts["insecure"], "-k")

    #     mock_msg.return_value.level = 5
    #     mock_msg.VER = 4
    #     geturl = GetURLexeCurl()
    #     geturl._set_defaults()
    #     self.assertEqual(geturl._opts["verbose"], "-v")
    #     self.assertEqual(geturl._files["url"], "")

    # def test_05__mkcurlcmd(self):
    #     """Test05 GetURLexeCurl()._mkcurlcmd()."""

    def test_06_get(self):
        """Test06 GetURLexeCurl().get()."""
        geturl = GetURLexeCurl()
        geturl._geturl = type('test', (object,), {})()
        geturl.get = self._get
        self.assertEqual(geturl.get("http://host"), "http://host")


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Management of execution mode"""

import os

from udocker.msg import Msg
from udocker.config import Config
from udocker.utils.fileutil import FileUtil
from udocker.utils.filebind import FileBind
from udocker.helper.elfpatcher import ElfPatcher
from udocker.engine.fakechroot import FakechrootEngine
from udocker.engine.proot import PRootEngine
from udocker.engine.runc import RuncEngine
from udocker.engine.singularity import SingularityEngine


class ExecutionMode(object):
    """Generic execution engine class to encapsulate the specific
    execution engines and their execution modes.
    P1: proot with seccomp
    P2: proot without seccomp (slower)
    F1: fakeroot running executables via direct loader invocation
    F2: similar to F1 with protected environment and modified ld.so
    F3: fakeroot patching the executables elf headers
    F4: similar to F3 with support to newly created executables
        dynamic patching of elf headers
    S1: singularity
    """

    def __init__(self, localrepo, container_id):
        self.localrepo = localrepo               # LocalRepository object
        self.container_id = container_id         # Container id
        self.container_dir = self.localrepo.cd_container(container_id)
        self.container_root = self.container_dir + "/ROOT"
        self.container_execmode = self.container_dir + "/execmode"
        self.container_orig_root = self.container_dir + "/root.path"
        self.exec_engine = None
        self.force_mode = None
        self.valid_modes = ("P1", "P2", "F1", "F2", "F3", "F4", "R1",
                            "R2", "R3", "S1")

    def get_mode(self):
        """Get execution mode"""
        if self.force_mode:
            return self.force_mode
        futil_xm = FileUtil(self.container_execmode)
        xmode = futil_xm.getdata('r').strip()
        if not xmode:
            xmode = Config.conf['default_execution_mode']
        return xmode

    def set_mode(self, xmode, force=False):
        """Set execution mode"""
        status = False
        prev_xmode = self.get_mode()
        elfpatcher = ElfPatcher(self.localrepo, self.container_id)
        filebind = FileBind(self.localrepo, self.container_id)
        futil_croot = FileUtil(self.container_orig_root)
        orig_path = futil_croot.getdata('r').strip()
        if xmode not in self.valid_modes:
            Msg().err("Error: invalid execmode:", xmode)
            return status

        if not (force or xmode != prev_xmode):
            return True

        if prev_xmode[0] in ('R', 'S') and xmode[0] != 'R':
            filebind.restore()

        if xmode[0] == 'F':
            if force or prev_xmode[0] in ("P", "R", "S"):
                futil_cont = FileUtil(self.container_root)
                status = (futil_cont.links_conv(force, True, orig_path)
                          and elfpatcher.get_ld_libdirs(force))

        if xmode in ('P1', 'P2', 'F1', 'R1', 'R2', 'R3', 'S1'):
            if prev_xmode in ('P1', 'P2', 'F1', 'R1', 'R2', 'R3', 'S1'):
                status = True
            elif force or prev_xmode in ("F2", "F3", "F4"):
                status = ((elfpatcher.restore_ld() or force) and
                          elfpatcher.restore_binaries())
            if xmode[0] == 'R':
                filebind.setup()
        elif xmode in ("F2", ):
            if force or prev_xmode in ("F3", "F4"):
                status = elfpatcher.restore_binaries()
            if force or prev_xmode in ('P1', 'P2', 'F1', 'R1',
                                       'R2', 'R3', 'S1'):
                status = elfpatcher.patch_ld()
        elif xmode in ("F3", "F4"):
            if force or prev_xmode in ('P1', 'P2', 'F1', 'F2',
                                       'R1', 'R2', 'R3', 'S1'):
                status = (elfpatcher.patch_ld() and
                          elfpatcher.patch_binaries())
            elif prev_xmode in ("F3", "F4"):
                status = True

        if xmode[0] in ("P", "R", "S"):
            if force or (status and prev_xmode.startswith("F")):
                futil = FileUtil(self.container_root)
                status = futil.links_conv(force, False, orig_path)

        if status or force:
            futil = FileUtil(self.container_execmode)
            status = futil.putdata(xmode, "w")

        if status or force:
            futil = FileUtil(self.container_orig_root)
            status = futil.putdata(os.path.realpath(self.container_root), "w")

        if (not status) and (not force):
            Msg().err("Error: container setup failed")

        return bool(status)

    def get_engine(self):
        """get execution engine instance"""
        xmode = self.get_mode()
        if xmode.startswith("P"):
            self.exec_engine = \
                PRootEngine(self.localrepo, self)
        elif xmode.startswith("F"):
            self.exec_engine = \
                FakechrootEngine(self.localrepo, self)
        elif xmode.startswith("R"):
            self.exec_engine = \
                RuncEngine(self.localrepo, self)
        elif xmode.startswith("S"):
            self.exec_engine = \
                SingularityEngine(self.localrepo, self)
        return self.exec_engine



# -*- coding: utf-8 -*-
"""Tools for udocker"""

import os
import sys
import tarfile
import random
import json
import stat

from udocker import is_genstr
from udocker.config import Config
from udocker.utils.curl import GetURL
from udocker.utils.fileutil import FileUtil
from udocker.msg import Msg
from udocker import __version__

def _str(data):
    """Safe str for Python 3 and Python 2"""
    if sys.version_info[0] >= 3:
        try:
            return data.decode()
        except (UnicodeDecodeError, AttributeError):
            pass
    return data


class UdockerTools(object):
    """Download and setup of the udocker supporting tools
    Includes: proot and alternative python modules, these
    are downloaded to facilitate the installation by the
    end-user.
    """

    def __init__(self, localrepo):
        self.localrepo = localrepo        # LocalRepository object
        self._autoinstall = Config.conf['autoinstall']  # True / False
        self._tarball = Config.conf['tarball']  # URL or file
        self._installinfo = Config.conf['installinfo']  # URL or file
        self._tarball_release = Config.conf['tarball_release']
        self._installretry = Config.conf['installretry']
        self._install_json = {}
        self.curl = GetURL()

    def _instructions(self):
        """
        Udocker installation instructions are available at:

          https://github.com/indigo-dc/udocker/tree/master/doc
          https://github.com/indigo-dc/udocker/tree/devel/doc

        Udocker requires additional tools to run. These are available
        in the udocker tarball. The tarballs are available at several
        locations. By default udocker will install from the locations
        defined in Config.tarball.

        To install from files or other URLs use these instructions:

        1) set UDOCKER_TARBALL to a remote URL or local filename:

          $ export UDOCKER_TARBALL=http://host/path
          or
          $ export UDOCKER_TARBALL=/tmp/filename

        2) run udocker with the install command or optionally using
           the option --force to overwrite the local installation:

          ./udocker install
          or
          ./udocker install --force

        3) once installed the binaries and containers will be placed
           by default under $HOME/.udocker

        """
        Msg().out(self._instructions.__doc__, l=Msg.ERR)
        Msg().out("udocker command line interface version:", __version__,
                  "\nrequires udockertools tarball release :",
                  self._tarball_release, l=Msg.ERR)

    def _version2int(self, version):
        """Convert version string to integer"""
        version_int = 0
        factor = 1000 * 1000
        for vitem in _str(version).strip().split('.'):
            try:
                version_int = version_int + (int(vitem) * factor)
            except (TypeError, ValueError):
                pass
            factor = factor / 1000
        return int(version_int)

    def _version_isok(self, version):
        """Is version >= than the minimum required tarball release"""
        if not (version and self._tarball_release):
            return False
        tarball_version_int = self._version2int(version)
        required_version_int = self._version2int(self._tarball_release)
        return tarball_version_int >= required_version_int

    def is_available(self):
        """Are the tools already installed"""
        version = \
            FileUtil(self.localrepo.libdir + "/VERSION").getdata('r').strip()
        return self._version_isok(version)

    def purge(self):
        """Remove existing files in bin, lib and doc"""
        for f_name in os.listdir(self.localrepo.bindir):
            f_path = self.localrepo.bindir + '/' + f_name
            FileUtil(f_path).register_prefix()
            FileUtil(f_path).remove(recursive=True)
        for f_name in os.listdir(self.localrepo.libdir):
            f_path = self.localrepo.libdir + '/' + f_name
            FileUtil(f_path).register_prefix()
            FileUtil(f_path).remove(recursive=True)
        for f_name in os.listdir(self.localrepo.docdir):
            f_path = self.localrepo.docdir + '/' + f_name
            FileUtil(f_path).register_prefix()
            FileUtil(f_path).remove(recursive=True)

    def _download(self, url):
        """Download a file """
        download_file = FileUtil("udockertools").mktmp()
        if Msg.level <= Msg.DEF:
            Msg().setlevel(Msg.NIL)
        (hdr, dummy) = self.curl.get(url, ofile=download_file, follow=True)
        if Msg.level == Msg.NIL:
            Msg().setlevel()
        try:
            if "200" in hdr.data["X-ND-HTTPSTATUS"]:
                return download_file
        except (KeyError, TypeError, AttributeError):
            pass
        FileUtil(download_file).remove()
        return ""

    def _get_file(self, url):
        """Get file from list of possible locations file or internet"""
        filename = ""
        if "://" in url:
            filename = self._download(url)
        elif os.path.exists(url):
            filename = os.path.realpath(url)
        if filename and os.path.isfile(filename):
            return filename
        return ""

    def _verify_version(self, tarball_file):
        """verify the tarball version"""
        if not (tarball_file and os.path.isfile(tarball_file)):
            return (False, "")

        tmpdir = FileUtil("VERSION").mktmpdir()
        if not tmpdir:
            return (False, "")

        try:
            tfile = tarfile.open(tarball_file, "r:gz")
            for tar_in in tfile.getmembers():
                if tar_in.name.startswith("udocker_dir/lib/VERSION"):
                    tar_in.name = os.path.basename(tar_in.name)
                    tfile.extract(tar_in, path=tmpdir)
            tfile.close()
        except tarfile.TarError:
            FileUtil(tmpdir).remove(recursive=True)
            return (False, "")
        tarball_version = FileUtil(tmpdir + "/VERSION").getdata('r').strip()
        status = self._version_isok(tarball_version)
        FileUtil(tmpdir).remove(recursive=True)
        return (status, tarball_version)

    def _install(self, tarball_file):
        """Install the tarball"""
        if not (tarball_file and os.path.isfile(tarball_file)):
            return False
        FileUtil(self.localrepo.topdir).chmod()
        self.localrepo.create_repo()

        try:
            tfile = tarfile.open(tarball_file, "r:gz")
            FileUtil(self.localrepo.bindir).rchmod()
            for tar_in in tfile.getmembers():
                if tar_in.name.startswith("udocker_dir/bin/"):
                    tar_in.name = os.path.basename(tar_in.name)
                    Msg().out("Info: extrating", tar_in.name, l=Msg.DBG)
                    tfile.extract(tar_in, self.localrepo.bindir)
            FileUtil(self.localrepo.bindir).rchmod(stat.S_IRUSR |
                                                   stat.S_IWUSR |
                                                   stat.S_IXUSR)

            FileUtil(self.localrepo.libdir).rchmod()
            for tar_in in tfile.getmembers():
                if tar_in.name.startswith("udocker_dir/lib/"):
                    tar_in.name = os.path.basename(tar_in.name)
                    Msg().out("Info: extrating", tar_in.name, l=Msg.DBG)
                    tfile.extract(tar_in, self.localrepo.libdir)
            FileUtil(self.localrepo.libdir).rchmod()

            FileUtil(self.localrepo.docdir).rchmod()
            for tar_in in tfile.getmembers():
                if tar_in.name.startswith("udocker_dir/doc/"):
                    tar_in.name = os.path.basename(tar_in.name)
                    Msg().out("Info: extrating", tar_in.name, l=Msg.DBG)
                    tfile.extract(tar_in, self.localrepo.docdir)
            FileUtil(self.localrepo.docdir).rchmod()
            tfile.close()
        except tarfile.TarError:
            return False
        return True

    def _get_mirrors(self, mirrors):
        """Get shuffled list of tarball mirrors"""
        if is_genstr(mirrors):
            mirrors = mirrors.split(' ')
        try:
            random.shuffle(mirrors)
        except NameError:
            pass
        return mirrors

    def get_installinfo(self):
        """Get json containing installation info"""
        Msg().out("Info: searching for messages:", l=Msg.VER)
        for url in self._get_mirrors(self._installinfo):
            infofile = self._get_file(url)
            try:
                with open(infofile, 'r', encoding='utf-8') as filep:
                    self._install_json = json.load(filep)
                for msg in self._install_json["messages"]:
                    Msg().out("Info:", msg)
            except (KeyError, AttributeError, ValueError,
                    OSError, IOError):
                Msg().out("Info: no messages:", infofile, url, l=Msg.VER)
            return self._install_json

    def _install_logic(self, force=False):
        """Obtain random mirror, download, verify and install"""
        for url in self._get_mirrors(self._tarball):
            Msg().out("Info: install using:", url, l=Msg.VER)
            tarballfile = self._get_file(url)
            (status, version) = self._verify_version(tarballfile)
            if status:
                Msg().out("Info: installing udockertools", version)
                status = self._install(tarballfile)
            elif force:
                Msg().out("Info: forcing install of udockertools", version)
                status = self._install(tarballfile)
            else:
                Msg().err("Error: version is", version, "for", url, l=Msg.VER)
            if "://" in url and tarballfile:
                FileUtil(tarballfile).remove()
            if status:
                return True
        return False

    def install(self, force=False):
        """Get the udocker tools tarball and install the binaries"""
        if self.is_available() and not force:
            return True

        if not self._autoinstall and not force:
            Msg().out("Warning: installation missing and autoinstall disabled",
                      l=Msg.WAR)
            return None

        if not self._tarball:
            Msg().out("Info: UDOCKER_TARBALL not set, installation skipped",
                      l=Msg.VER)
            return True

        Msg().out("Info: udocker command line interface", __version__)
        Msg().out("Info: searching for udockertools",
                  self._tarball_release, l=Msg.INF)
        retry = self._installretry
        while  retry:
            if self._install_logic(force):
                self.get_installinfo()
                Msg().out("Info: installation of udockertools successful")
                return True

            retry = retry - 1
            Msg().err("Error: installation failure retrying ...", l=Msg.VER)

        self._instructions()
        self.get_installinfo()
        Msg().err("Error: installation of udockertools failed")
        return False



#!/usr/bin/env python
"""
udocker unit tests: Unique
"""

from unittest import TestCase, main
from udocker.helper.unique import Unique


class UniqueTestCase(TestCase):
    """Test Unique() class."""

    def test_01_init(self):
        """Test01 Unique() constructor."""
        uniq = Unique()
        self.assertEqual(uniq.def_name, "udocker")

    def test_02__rnd(self):
        """Test02 Unique._rnd()."""
        uniq = Unique()
        rand = uniq._rnd(64)
        self.assertIsInstance(rand, str)
        self.assertEqual(len(rand), 64)

    def test_03_uuid(self):
        """Test03 Unique.uuid()."""
        uniq = Unique()
        rand = uniq.uuid("zxcvbnm")
        self.assertEqual(len(rand), 36)
        rand = uniq.uuid(789)
        self.assertEqual(len(rand), 36)

    def test_04_imagename(self):
        """Test04 Unique.imagename()."""
        uniq = Unique()
        rand = uniq.imagename()
        self.assertEqual(len(rand), 16)

    def test_05_imagetag(self):
        """Test05 Unique.imagetag()."""
        uniq = Unique()
        rand = uniq.imagetag()
        self.assertEqual(len(rand), 10)

    def test_06_layer_v1(self):
        """Test06 Unique.layer_v1()."""
        uniq = Unique()
        rand = uniq.layer_v1()
        self.assertEqual(len(rand), 64)

    def test_07_filename(self):
        """Test07 Unique.filename()."""
        uniq = Unique()
        rand = uniq.filename("zxcvbnmasdf")
        self.assertTrue(rand.endswith("zxcvbnmasdf"))
        self.assertTrue(rand.startswith("udocker"))
        self.assertGreater(len(rand.strip()), 56)
        rand = uniq.filename(12345)
        self.assertTrue(rand.endswith("12345"))
        self.assertTrue(rand.startswith("udocker"))
        self.assertGreater(len(rand.strip()), 50)


if __name__ == '__main__':
    main()
